from dataclasses import MISSING, dataclass
from pathlib import Path
from typing import Optional

import yaml


@dataclass(frozen=True)
class Setting:
    root_path: Path = Path("~/.mohou")
    primary_project_name: Optional[str] = None

    @classmethod
    def construct(cls) -> "Setting":

        default_attributes = {}
        for key, val in cls.__dataclass_fields__.items():  # type: ignore # mypy's bug
            assert val.default != MISSING
            default_attributes[key] = val.default

        setting = None
        setting_path = Path("~/.mohou").expanduser() / "setting.yaml"
        if setting_path.exists():
            with setting_path.open(mode="r") as f:
                setting = yaml.safe_load(f)

        if setting is None:
            return cls()  # use default

        invalid_keys = set(setting.keys()).difference(set(default_attributes.keys()))
        message = "invalid keys {} in setting.yaml".format(invalid_keys)
        assert len(invalid_keys) == 0, message

        for key in default_attributes.keys():
            if key in setting:
                # overwrite
                raw_type = cls.__dataclass_fields__[key].type  # type: ignore # mypy's bug
                if isinstance(raw_type, type):
                    expect_type = raw_type
                else:
                    is_union_t = len(raw_type.__args__) == 2
                    is_optional_t = is_union_t and issubclass(raw_type.__args__[1], type(None))
                    assert is_optional_t
                    expect_type = raw_type.__args__[0]  # the one that is not NoneType
                default_attributes[key] = expect_type(setting[key])  # cast

        return cls(**default_attributes)  # type: ignore

    def __post_init__(self):
        for key, val in self.__dict__.items():
            if isinstance(val, Path):
                self.__dict__[key] = val.expanduser()


setting = Setting.construct()



import base64
import pickle
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, Generic, List, Optional, Tuple, Type, TypeVar

import numpy as np
import torch
from sklearn.decomposition import PCA

from mohou.model.autoencoder import AutoEncoderBase
from mohou.model.common import ModelBase
from mohou.types import ElementT, EpisodeBundle, ImageT, VectorT, get_element_type
from mohou.utils import assert_equal_with_message, assert_isinstance_with_message

EncoderT = TypeVar("EncoderT", bound="EncoderBase")


@dataclass  # type: ignore
class EncoderBase(ABC, Generic[ElementT]):
    elem_type: Type[ElementT]
    input_shape: Tuple[int, ...]
    output_size: int

    def to_dict(self) -> Dict:
        d: Dict = {}
        d["elem_type"] = self.elem_type.__name__
        d["input_shape"] = self.input_shape
        d["output_size"] = self.output_size
        # TODO: automate
        return d

    @classmethod
    def from_dict(cls: Type[EncoderT], d: Dict) -> EncoderT:
        d["elem_type"] = get_element_type(d["elem_type"])
        d["input_shape"] = tuple(d["input_shape"])
        return cls(**d)

    def forward(self, inp: ElementT, check_size: bool = True) -> np.ndarray:
        assert_isinstance_with_message(inp, self.elem_type)

        if check_size:
            assert_equal_with_message(inp.shape, self.input_shape, "input shape")

        out = self._forward_impl(inp)

        if check_size:
            assert_equal_with_message(out.shape, (self.output_size,), "output shape")

        return out

    def backward(self, inp: np.ndarray, check_size: bool = True) -> ElementT:
        if check_size:
            assert_equal_with_message(inp.shape, (self.output_size,), "input shape")

        out = self._backward_impl(inp)

        if check_size:
            assert_equal_with_message(out.shape, self.input_shape, "input shape")

        return out

    @abstractmethod
    def _forward_impl(self, inp: ElementT) -> np.ndarray:
        pass

    @abstractmethod
    def _backward_impl(self, inp: np.ndarray) -> ElementT:
        pass

    def __eq__(self, other: object) -> bool:
        # DO NOT USE dataclass's default __eq__
        if not isinstance(other, EncoderBase):
            return NotImplemented
        assert type(self) == type(other)
        for attr in ["elem_type", "input_shape", "output_size"]:
            if not self.__dict__[attr] == other.__dict__[attr]:
                return False
        return True


class HasAModel(ABC):
    @abstractmethod
    def _get_model(self) -> ModelBase:
        pass

    def get_device(self) -> Optional[torch.device]:
        return self._get_model().device

    def set_device(self, device: torch.device) -> None:
        self._get_model().put_on_device(device)


@dataclass(eq=False)
class ImageEncoder(EncoderBase[ImageT], HasAModel):
    input_shape: Tuple[int, int, int]
    model: AutoEncoderBase

    def __post_init__(self):
        inp_dummy = self.elem_type.dummy_from_shape(self.input_shape[:2])
        out_dummy = self._forward_impl(inp_dummy)
        assert_equal_with_message(out_dummy.shape, (self.output_size,), "shape")

        inp_reconstucted = self._backward_impl(out_dummy)
        assert_isinstance_with_message(inp_reconstucted, self.elem_type)

    @classmethod
    def from_auto_encoder(cls, model: AutoEncoderBase) -> "ImageEncoder":
        image_type = model.image_type
        np_image_shape = (model.config.n_pixel, model.config.n_pixel, model.channel())
        return cls(image_type, np_image_shape, model.config.n_bottleneck, model)

    def to_dict(self) -> Dict:
        d = super().to_dict()
        d["model"] = base64.b64encode(pickle.dumps(self.model)).decode("utf-8")
        return d

    @classmethod
    def from_dict(cls, d: Dict) -> "ImageEncoder":
        d["model"] = pickle.loads(base64.b64decode(d["model"].encode()))
        return super().from_dict(d)

    def _forward_impl(self, inp: ImageT) -> np.ndarray:
        inp_tensor = inp.to_tensor().unsqueeze(dim=0).to(self.get_device())
        out_tensor = self.model.encode(inp_tensor).squeeze(dim=0)
        out_numpy = out_tensor.cpu().detach().numpy()
        return out_numpy

    def _backward_impl(self, inp: np.ndarray) -> ImageT:
        inp_tensor = torch.from_numpy(inp).unsqueeze(dim=0).float()
        inp_tensor = inp_tensor.to(self.get_device())
        out_tensor = self.model.decode(inp_tensor).squeeze(dim=0).cpu()
        out: ImageT = self.elem_type.from_tensor(out_tensor)
        return out

    def __eq__(self, other: object) -> bool:
        # DO NOT USE dataclass's default __eq__
        if not isinstance(other, ImageEncoder):
            return NotImplemented
        assert type(self) == type(other)

        if not super().__eq__(other):
            return False

        # https://discuss.pytorch.org/t/check-if-models-have-same-weights/4351/2
        model1 = self.model
        model2 = other.model
        for p1, p2 in zip(model1.parameters(), model2.parameters()):
            if p1.data.ne(p2.data).sum() > 0:
                return False
        return True

    def _get_model(self) -> ModelBase:
        return self.model


@dataclass
class VectorIdenticalEncoder(EncoderBase[VectorT]):
    input_shape: Tuple[int]

    @classmethod
    def create(cls, vector_type: Type[VectorT], dimension: int) -> "VectorIdenticalEncoder":
        return cls(vector_type, (dimension,), dimension)

    def _forward_impl(self, inp: VectorT) -> np.ndarray:
        return inp.numpy()

    def _backward_impl(self, inp: np.ndarray) -> VectorT:
        return self.elem_type(inp)


class VectorPCAEncoder(EncoderBase[VectorT]):
    input_shape: Tuple[int]
    pca: PCA

    def __init__(self, vector_type: Type[VectorT], pca: PCA):
        input_shape = (pca.n_features_,)
        output_dimension = pca.n_components
        super().__init__(vector_type, input_shape, output_dimension)

        self.pca = pca

    def to_dict(self) -> Dict:  # type: ignore
        assert False, "currently not supported"

    @classmethod
    def from_dict(cls, d: Dict) -> "ImageEncoder":  # type: ignore
        assert False, "currently not supported"

    def _forward_impl(self, inp: VectorT) -> np.ndarray:
        inp_as_2d = np.expand_dims(inp.numpy(), axis=0)
        out = self.pca.transform(inp_as_2d)
        return out.flatten()

    def _backward_impl(self, inp: np.ndarray) -> VectorT:
        out = self.pca.inverse_transform(np.expand_dims(inp, axis=0))
        return self.elem_type(out.flatten())

    @classmethod
    def from_bundle(
        cls, bundle: EpisodeBundle, vector_type: Type[VectorT], n_out: int
    ) -> "VectorPCAEncoder[VectorT]":
        elem_list: List[VectorT] = []
        for episode in bundle.get_touch_bundle():
            elem_seq = episode.get_sequence_by_type(vector_type)
            elem_list.extend(elem_seq.elem_list)
        mat = np.array([e.numpy() for e in elem_list])
        pca = PCA(n_components=n_out)
        pca.fit(mat)
        return cls(vector_type, pca)



import copy
import logging
from dataclasses import dataclass
from typing import Generic, List, Optional, Type

import torch
from torch.utils.data import Dataset

from mohou.types import EpisodeBundle, ImageT

logger = logging.getLogger(__name__)


@dataclass
class AutoEncoderDatasetConfig:
    batch_augment_factor: int = 2  # if you have large enough RAM, set to large (like 4)

    def __post_init__(self):
        assert self.batch_augment_factor >= 0
        logger.info("autoencoder dataset config: {}".format(self))


@dataclass
class AutoEncoderDataset(Dataset, Generic[ImageT]):
    image_type: Type[ImageT]
    image_list: List[ImageT]

    def __len__(self) -> int:
        return len(self.image_list)

    def __getitem__(self, idx) -> torch.Tensor:
        return self.image_list[idx].to_tensor()

    @classmethod
    def from_bundle(
        cls,
        bundle: EpisodeBundle,
        image_type: Type[ImageT],
        augconfig: Optional[AutoEncoderDatasetConfig] = None,
    ) -> "AutoEncoderDataset":

        if augconfig is None:
            augconfig = AutoEncoderDatasetConfig()

        image_list: List[ImageT] = []
        for episode_data in bundle:
            image_list.extend(episode_data.get_sequence_by_type(image_type))

        image_list_rand = copy.deepcopy(image_list)
        for i in range(augconfig.batch_augment_factor):
            image_list_rand.extend([copy.deepcopy(image).randomize() for image in image_list])

        return cls(image_type, image_list_rand)



import logging
import random
import time
from logging import Logger
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type

import matplotlib.pyplot as plt
import numpy as np
import tqdm

from mohou.model.autoencoder import VariationalAutoEncoder

try:
    from moviepy.editor import ImageSequenceClip
except Exception:
    ImageSequenceClip = None

from mohou.dataset import (
    AutoEncoderDataset,
    AutoEncoderDatasetConfig,
    AutoRegressiveDataset,
    AutoRegressiveDatasetConfig,
)
from mohou.default import load_default_image_encoder
from mohou.encoding_rule import EncodingRule, EncodingRuleBase
from mohou.model import (
    LSTM,
    PBLSTM,
    AutoEncoder,
    AutoEncoderBase,
    AutoEncoderConfig,
    LSTMConfig,
    PBLSTMConfig,
)
from mohou.model.lstm import LSTMBase, LSTMConfigBase
from mohou.propagator import PropagatorBase
from mohou.trainer import TrainCache, TrainConfig, train
from mohou.types import (
    ElementBase,
    ElementDict,
    EpisodeBundle,
    EpisodeData,
    ImageBase,
    TerminateFlag,
)
from mohou.utils import canvas_to_ndarray

logger = logging.getLogger(__name__)


def create_default_logger(project_path: Path, prefix: str) -> Logger:
    timestr = "_" + time.strftime("%Y%m%d%H%M%S")
    log_dir_path = project_path / "log"
    log_dir_path.mkdir(parents=True, exist_ok=True)
    log_file_path = log_dir_path / (prefix + timestr + ".log")

    FORMAT = "[%(levelname)s] %(asctime)s %(name)s: %(message)s"
    logging.basicConfig(filename=str(log_file_path), format=FORMAT)
    logger = logging.getLogger("mohou")
    logger.setLevel(level=logging.INFO)

    log_sym_path = log_dir_path / ("latest_" + prefix + ".log")

    logger.info("create log symlink :{0} => {1}".format(log_file_path, log_sym_path))
    if log_sym_path.is_symlink():
        log_sym_path.unlink()
    log_sym_path.symlink_to(log_file_path)

    return logger


def train_autoencoder(
    project_path: Path,
    image_type: Type[ImageBase],
    model_config: AutoEncoderConfig,
    dataset_config: AutoEncoderDatasetConfig,
    train_config: TrainConfig,
    ae_type: Type[AutoEncoderBase] = AutoEncoder,
    bundle: Optional[EpisodeBundle] = None,
    warm_start: bool = False,
):

    if bundle is None:
        bundle = EpisodeBundle.load(project_path)

    dataset = AutoEncoderDataset.from_bundle(bundle, image_type, dataset_config)
    if warm_start:
        logger.info("warm start")
        tcache = TrainCache.load(project_path, ae_type)
        train(project_path, tcache, dataset, config=train_config)
    else:
        model = ae_type(model_config)  # type: ignore
        tcache = TrainCache.from_model(model)  # type: ignore[var-annotated]
        train(project_path, tcache, dataset, config=train_config)


def train_lstm(
    project_path: Path,
    encoding_rule: EncodingRule,
    model_config: LSTMConfigBase,
    dataset_config: AutoRegressiveDatasetConfig,
    train_config: TrainConfig,
    model_type: Type[LSTMBase] = LSTM,
    bundle: Optional[EpisodeBundle] = None,
    warm_start: bool = False,
    context_list: Optional[List[np.ndarray]] = None,
):

    assert model_config.window_size == dataset_config.window_size
    if model_config.window_size is not None:
        logger.warning(
            "NOTE(experimental): lstm with window size may be dropped without any notification"
        )

    # a dirty assertion TODO: do this by generic typing
    compat_table: Dict[Type[LSTMBase], Type[LSTMConfigBase]] = {
        LSTM: LSTMConfig,
        PBLSTM: PBLSTMConfig,
    }
    assert model_type in compat_table
    assert compat_table[model_type] == type(model_config)

    if bundle is None:
        bundle = EpisodeBundle.load(project_path)

    if context_list is None:
        assert model_config.n_static_context == 0
    else:
        for context in context_list:
            assert len(context) == model_config.n_static_context

    dataset = AutoRegressiveDataset.from_bundle(
        bundle,
        encoding_rule,
        dataset_config=dataset_config,
        static_context_list=context_list,
    )

    if warm_start:
        logger.info("warm start")
        tcache = TrainCache.load(project_path, LSTM)
        train(project_path, tcache, dataset, config=train_config)
    else:
        model = model_type(model_config)
        tcache = TrainCache.from_model(model)  # type: ignore
        train(project_path, tcache, dataset, config=train_config)


def visualize_train_histories(project_path: Path):
    plot_dir_path = project_path / "train_history"
    plot_dir_path.mkdir(exist_ok=True)

    all_result_paths = TrainCache.filter_result_paths(project_path, None)
    for result_path in all_result_paths:
        tcache = TrainCache.load_from_cache_path(result_path)
        image_path = plot_dir_path / (result_path.name + ".png")
        fig, ax = tcache.visualize()
        fig.savefig(str(image_path))
        print("saved to {}".format(image_path))


def visualize_image_reconstruction(
    project_path: Path,
    bundle: EpisodeBundle,
    autoencoder: AutoEncoderBase,
    n_vis: int = 5,
    prefix: Optional[str] = None,
):
    save_dir_path = project_path / "autoencoder_result"
    save_dir_path.mkdir(exist_ok=True)

    bundle_untouch = bundle.get_untouch_bundle()
    bundle_touch = bundle.get_touch_bundle()

    image_type = autoencoder.image_type  # type: ignore[union-attr]
    no_aug = AutoEncoderDatasetConfig(0)  # to feed not randomized image
    dataset_untouch = AutoEncoderDataset.from_bundle(bundle_untouch, image_type, no_aug)
    dataset_touch = AutoEncoderDataset.from_bundle(bundle_touch, image_type, no_aug)

    for dataset, postfix in zip([dataset_untouch, dataset_touch], ["untouch", "touch"]):
        idxes = list(range(len(dataset)))
        random.shuffle(idxes)
        idxes_test = idxes[: min(n_vis, len(dataset))]

        for i, idx in enumerate(idxes_test):

            image_torch = dataset[idx].unsqueeze(dim=0)
            image_torch_reconstructed = autoencoder(image_torch)

            img = dataset.image_type.from_tensor(image_torch.squeeze(dim=0))
            img_reconstructed = dataset.image_type.from_tensor(
                image_torch_reconstructed.squeeze(dim=0)
            )

            fig, (ax1, ax2) = plt.subplots(1, 2)
            fig.suptitle("left: original, right: reconstructed")
            ax1.imshow(img.to_rgb().numpy())
            ax2.imshow(img_reconstructed.to_rgb().numpy())
            filename = "result-{}-{}.png".format(postfix, i)
            if prefix is not None:
                filename = prefix + "-" + str(filename)
            file_path = save_dir_path / filename
            plt.savefig(str(file_path))


def visualize_variational_autoencoder(project_path: Path) -> None:

    tcache = TrainCache[VariationalAutoEncoder[ImageBase]].load(
        project_path, VariationalAutoEncoder
    )
    vae = tcache.best_model

    save_dir_path = project_path / "autoencoder_result"
    save_dir_path.mkdir(exist_ok=True)

    for axis in range(vae.config.n_bottleneck):
        images = vae.get_latent_axis_images(axis)
        assert ImageSequenceClip is not None, "check if your moviepy is properly installed"
        clip = ImageSequenceClip([im.to_rgb().numpy() for im in images], fps=20)
        file_path = save_dir_path / "vae-axis{}.gif".format(axis)
        clip.write_gif(str(file_path), fps=20)


def add_text_to_image(image: ImageBase, text: str, color: str):
    fig = plt.figure(tight_layout={"pad": 0})
    ax = plt.subplot(1, 1, 1)
    ax.axis("off")
    ax.imshow(image.to_rgb().numpy())
    bbox = dict(boxstyle="round", facecolor="white", alpha=0.4)
    ax.text(7, 1, text, fontsize=15, color=color, verticalalignment="top", bbox=bbox)
    fig.canvas.draw()
    fig.canvas.flush_events()
    return canvas_to_ndarray(fig)


class VectorSequencePlotter:
    fig: Any
    axs: List
    axis_index_table: Dict[Tuple[Type[ElementBase], int], int]
    inverse_axis_index_table: List[Tuple[Type[ElementBase], int]]
    min_val_per_axis: List[float]
    max_val_per_axis: List[float]
    is_finalized: bool = False

    def __init__(self, encoding_rule: EncodingRuleBase) -> None:

        axis_index_table: Dict[Tuple[Type[ElementBase], int], int] = {}
        inverse_axis_index_table: List[Tuple[Type[ElementBase], int]] = []
        axis_count = 0
        for elem_type, encoder in encoding_rule.items():
            for i in range(encoder.output_size):
                axis_index_table[(elem_type, i)] = axis_count
                inverse_axis_index_table.append((elem_type, i))
                axis_count += 1

        fig, axs = plt.subplots(axis_count, figsize=(5, 0.5 * axis_count))
        if axis_count == 1:
            axs = [axs]

        for ax in axs:
            ax.set_xticklabels([])
            ax.yaxis.set_tick_params(labelsize=5)

        fig.tight_layout(pad=3.0)

        self.fig = fig
        self.axs = axs
        self.axis_index_table = axis_index_table
        self.inverse_axis_index_table = inverse_axis_index_table
        self.min_val_per_axis = [np.inf for _ in range(axis_count)]
        self.max_val_per_axis = [-np.inf for _ in range(axis_count)]

    def add_plot(
        self,
        x_seq: Optional[np.ndarray],
        y_seq: np.ndarray,
        elem_type: Type[ElementBase],
        idx: int,
        **kwargs
    ):
        assert y_seq.ndim == 1
        assert not self.is_finalized
        axis_index = self.axis_index_table[(elem_type, idx)]
        if x_seq is not None:
            assert x_seq.ndim == 1
            self.axs[axis_index].plot(x_seq, y_seq, **kwargs)
        else:
            self.axs[axis_index].plot(y_seq, **kwargs)
        self.min_val_per_axis[axis_index] = min(self.min_val_per_axis[axis_index], np.min(y_seq))
        self.max_val_per_axis[axis_index] = max(self.max_val_per_axis[axis_index], np.max(y_seq))

    def finalize(self):
        ax_last = self.axs[-1]
        ax_last.legend(fontsize=5)

        for axis_index, ax in enumerate(self.axs):
            val_min = self.min_val_per_axis[axis_index]
            val_max = self.max_val_per_axis[axis_index]
            margin = max(0.1, 0.1 * (val_max - val_min))
            ax.set_ylim(val_min - margin, val_max + margin)
            ax.grid()

            elem_type, vector_index = self.inverse_axis_index_table[axis_index]
            ax.text(
                0.5,
                0.5,
                "{} dimension {}".format(elem_type.__name__, vector_index),
                fontsize=5,
                transform=ax.transAxes,
                horizontalalignment="center",
                verticalalignment="center",
                bbox=dict(boxstyle="round", facecolor="white", alpha=0.7),
            )
        self.is_finalized = True

    def save_fig(self, image_path: Path):
        assert self.is_finalized
        self.fig.savefig(str(image_path), format="png", dpi=300, bbox_inches="tight")


def visualize_lstm_propagation(project_path: Path, propagator: PropagatorBase, n_prop: int):
    bundle = EpisodeBundle.load(project_path).get_untouch_bundle()
    save_dir_path = project_path / "lstm_result"
    save_dir_path.mkdir(exist_ok=True)
    prop_name = propagator.__class__.__name__

    if propagator.require_static_context:
        raise NotImplementedError("not implemented yet")

    image_encoder = load_default_image_encoder(project_path)
    image_type = image_encoder.elem_type

    for idx, episode_data in enumerate(bundle):

        print("start lstm propagation")
        n_feed = 20

        edict_feed_pred = []
        for i in range(n_feed):
            edict = episode_data[i]
            edict_required = edict.get_subdict(propagator.encoding_rule.keys())
            propagator.feed(edict_required)
            edict_feed_pred.append(edict_required)

        pred_edict_list = propagator.predict(n_prop)
        edict_feed_pred.extend(pred_edict_list)

        episode_feedpred = EpisodeData.from_edict_list(edict_feed_pred, check_terminate_flag=False)
        print("finish lstm propagation")

        encoding_rule = propagator.encoding_rule
        arr_gtruth = encoding_rule.apply_to_episode_data(episode_data)
        arr_feedpred = encoding_rule.apply_to_episode_data(episode_feedpred)

        vecseq_plotter = VectorSequencePlotter(encoding_rule)

        for elem_type, bound in encoding_rule.type_bound_table.items():
            assert bound.step is None
            arr_partial_gt = arr_gtruth[:, bound.start : bound.stop]
            arr_partial_fp = arr_feedpred[:, bound.start : bound.stop]

            n_dim = bound.stop - bound.start
            for i in range(n_dim):
                y_seq_gt = arr_partial_gt[:, i]
                vecseq_plotter.add_plot(
                    None, y_seq_gt, elem_type, i, color="blue", lw=1, label="ground truth"
                )
                y_seq_fp = arr_partial_fp[:, i]
                vecseq_plotter.add_plot(
                    None, y_seq_fp, elem_type, i, color="red", lw=1, label="feed + pred"
                )

        image_path = save_dir_path / "seq-vectors-{}{}.png".format(prop_name, idx)
        vecseq_plotter.finalize()
        vecseq_plotter.save_fig(image_path)
        print("saved to {}".format(image_path))

        # save gif image
        print("adding text to images...")
        images = episode_feedpred.get_sequence_by_type(image_type)
        flags = episode_feedpred.get_sequence_by_type(TerminateFlag)
        feed_images = images[:n_feed]
        pred_images = images[n_feed:]
        pred_flags = flags[n_feed:]

        fed_images_with_text = [
            add_text_to_image(image, "fed (original) image)", "blue") for image in feed_images
        ]
        clamp = lambda x: max(min(x, 1.0), 0.0)  # noqa
        pred_images_with_text = [
            add_text_to_image(
                image,
                "predicted image (prob-terminated={:.2f})".format(clamp(flag.numpy().item())),
                "green",
            )
            for image, flag in zip(pred_images, pred_flags)
        ]

        images_with_text = fed_images_with_text + pred_images_with_text

        image_path = save_dir_path / "seq-{}-rgb{}.gif".format(prop_name, idx)
        assert ImageSequenceClip is not None, "check if your moviepy is properly installed"
        clip = ImageSequenceClip(images_with_text, fps=20)
        clip.write_gif(str(image_path), fps=20)


def plot_execution_result(
    project_path: Path, propagator: PropagatorBase, edict_seq: List[ElementDict], n_prop: int = 10
):
    timestr = "_" + time.strftime("%Y%m%d%H%M%S")
    propagator.reset()

    # plot actual data
    episode = EpisodeData.from_edict_list(edict_seq, check_terminate_flag=False)
    encoding_rule = propagator.encoding_rule
    vecseq_plotter = VectorSequencePlotter(encoding_rule)
    arr = encoding_rule.apply_to_episode_data(episode)

    for elem_type, bound in encoding_rule.type_bound_table.items():
        assert bound.step is None
        arr_partial = arr[:, bound.start : bound.stop]
        n_dim = bound.stop - bound.start
        for i in range(n_dim):
            y_seq = arr_partial[:, i]
            vecseq_plotter.add_plot(None, y_seq, elem_type, i, color="blue", lw=1.0)

    # plot prediction
    for time_index, edict in enumerate(tqdm.tqdm(edict_seq)):
        x_seq = np.array([time_index + 1 + j for j in range(n_prop)])

        propagator.feed(edict)
        edict_list_predicted = propagator.predict(n_prop)
        episode_predicted = EpisodeData.from_edict_list(
            edict_list_predicted, check_terminate_flag=False
        )
        arr = encoding_rule.apply_to_episode_data(episode_predicted)

        for elem_type, bound in encoding_rule.type_bound_table.items():
            assert bound.step is None
            arr_partial = arr[:, bound.start : bound.stop]
            n_dim = bound.stop - bound.start
            for i in range(n_dim):
                y_seq = arr_partial[:, i]
                vecseq_plotter.add_plot(
                    x_seq, y_seq, elem_type, i, color="red", lw=0.1, marker=".", markersize="0.2"
                )

    save_dir_path = project_path / "execution_result"
    save_dir_path.mkdir(exist_ok=True)
    prop_name = propagator.__class__.__name__
    image_path = save_dir_path / "result-{}-{}.png".format(prop_name, timestr)
    vecseq_plotter.finalize()
    vecseq_plotter.save_fig(image_path)



# flake8: noqa
from mohou.dataset.autoencoder_dataset import (
    AutoEncoderDataset,
    AutoEncoderDatasetConfig,
)
from mohou.dataset.chimera_dataset import ChimeraDataset
from mohou.dataset.sequence_dataset import (
    AutoRegressiveDataset,
    AutoRegressiveDatasetConfig,
    MarkovControlSystemDataset,
    SequenceDatasetConfig,
)



import logging
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Type, TypeVar, Union

import torch
import torch.nn as nn
from torch.nn.parameter import Parameter

from mohou.model.common import LossDict, ModelBase, ModelConfigBase
from mohou.model.third_party import VariationalLSTM
from mohou.types import ElementBase

logger = logging.getLogger(__name__)


@dataclass
class LSTMConfigBase(ModelConfigBase):
    n_state_with_flag: int
    n_static_context: int = 0
    n_hidden: int = 200
    n_layer: int = 4
    n_output_layer: int = 1
    window_size: Optional[int] = None
    variational: bool = False


LSTMConfigBaseT = TypeVar("LSTMConfigBaseT", bound=LSTMConfigBase)


@dataclass
class LSTMConfig(LSTMConfigBase):
    """
    type_wise_loss: if True, loss is computed for each type (following type_bound_table) and
    each loss is stored in the LossDict
    """

    type_wise_loss: bool = False
    type_bound_table: Optional[Dict[Type[ElementBase], slice]] = None

    def __post_init__(self):
        if self.type_wise_loss:
            assert self.type_bound_table is not None


class LSTMBase(ModelBase[LSTMConfigBaseT]):
    @staticmethod
    def _setup_inner(
        n_input: int,
        n_output: int,
        n_hidden: int,
        n_layer: int,
        n_output_layer: int,
        variational: bool,
    ) -> Tuple[Union[nn.LSTM, VariationalLSTM], nn.Sequential]:

        if variational:
            lstm_layer: Union[nn.LSTM, VariationalLSTM] = VariationalLSTM(
                n_input, n_hidden, n_layer, dropouti=0.5, dropoutw=0.5, dropouto=0.5
            )
        else:
            lstm_layer = nn.LSTM(n_input, n_hidden, n_layer, batch_first=True)

        output_layers = []
        for _ in range(n_output_layer):
            output_layers.append(nn.Linear(n_hidden, n_hidden))
        output_layers.append(nn.Linear(n_hidden, n_output))
        output_layer = nn.Sequential(*output_layers)
        return lstm_layer, output_layer

    @staticmethod
    def _loss_inner_type_wise(
        state_output_ref: torch.Tensor,
        state_output_pred: torch.Tensor,
        type_bound_table: Dict[Type[ElementBase], slice],
    ) -> LossDict:
        # NOTE: This is an experimental feature. Compute type-wise prediction loss. LossDict looks like
        # d["AngleVector"] = 0.002, d["RGBImage"] = 0.0003

        def bound2dim(bound: slice) -> int:
            return bound.stop - bound.start

        total_dim = sum([bound2dim(bound) for bound in type_bound_table.values()])

        d = {}
        for elem_type, bound in type_bound_table.items():

            scaler = bound2dim(bound) / total_dim

            pred_typewise = state_output_pred[:, :, bound]
            state_sample_output_typewise = state_output_ref[:, :, bound]
            loss_value_partial = nn.MSELoss()(pred_typewise, state_sample_output_typewise) * scaler
            key = elem_type.__name__
            d[key] = loss_value_partial

        # take average
        return LossDict(d)

    @staticmethod
    def _loss_inner(state_output_ref: torch.Tensor, state_output_pred: torch.Tensor) -> LossDict:
        loss_value = nn.MSELoss()(state_output_pred, state_output_ref)
        return LossDict({"prediction": loss_value})

    def _sanity_check_context_sample(
        self, state_sample: torch.Tensor, context_sample: torch.Tensor
    ) -> None:
        n_batch, n_seq_len, n_dim = state_sample.shape
        assert context_sample.ndim == 2
        assert context_sample.shape[0] == n_batch
        assert context_sample.shape[1] == self.config.n_static_context


LSTMBaseT = TypeVar("LSTMBaseT", bound=LSTMBase)


class LSTM(LSTMBase[LSTMConfig]):
    """
    lstm: x_t+1 = f(x_t, x_t-1, ...)
    lstm with context: x_t+1 = f(x_t, x_t-1, ..., c) where c is static (time-invariant) context
    """

    lstm_layer: Union[nn.LSTM, VariationalLSTM]
    output_layer: nn.Sequential

    def _setup_from_config(self, config: LSTMConfig) -> None:
        n_input = config.n_state_with_flag + config.n_static_context
        n_output = config.n_state_with_flag
        self.lstm_layer, self.output_layer = self._setup_inner(
            n_input,
            n_output,
            config.n_hidden,
            config.n_layer,
            config.n_output_layer,
            config.variational,
        )

    def loss(self, sample: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]) -> LossDict:

        _, state_sample, static_context_sample = sample
        self._sanity_check_context_sample(state_sample, static_context_sample)

        # propagation
        state_sample_input, state_sample_output = state_sample[:, :-1], state_sample[:, 1:]
        pred, _ = self.forward(state_sample_input, static_context_sample)

        if self.config.type_wise_loss:
            assert self.config.type_bound_table is not None
            return self._loss_inner_type_wise(
                state_sample_output, pred, self.config.type_bound_table
            )
        else:
            return self._loss_inner(state_sample_output, pred)

    def forward(
        self,
        state_sample: torch.Tensor,
        context_sample: torch.Tensor,
        hidden: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        self._sanity_check_context_sample(state_sample, context_sample)

        # arrange bais_sample and create concat state_sample
        _, n_seqlen, _ = state_sample.shape

        if self.config.window_size is not None:
            assert hidden is None, "not tested for this case"
            if n_seqlen > self.config.window_size:
                n_seqlen_new = min(self.config.window_size, n_seqlen)
                message = "warn: use only recent {} states of given {} states".format(
                    n_seqlen_new, n_seqlen
                )
                logger.warning(message)
                state_sample = state_sample[:, -n_seqlen_new:]
                n_seqlen = n_seqlen_new

        context_unsqueezed = context_sample.unsqueeze(dim=1)
        context_sequenced = context_unsqueezed.expand(-1, n_seqlen, -1)
        context_auged_state_sample = torch.cat((state_sample, context_sequenced), dim=2)

        # propagation
        preout, hidden = self.lstm_layer(context_auged_state_sample, hidden)
        out = self.output_layer(preout)
        assert hidden is not None
        return out, hidden


@dataclass
class PBLSTMConfig(LSTMConfigBase):
    # TODO: add static_contex, type_wise_loss
    n_pb: int = -1  # override this
    n_pb_dim: int = 2


class PBLSTM(LSTMBase[PBLSTMConfig]):
    lstm_layer: Union[nn.LSTM, VariationalLSTM]
    output_layer: nn.Sequential
    parametric_bias_list: List[Parameter]

    def _setup_from_config(self, config: PBLSTMConfig) -> None:
        n_input = config.n_state_with_flag + config.n_pb_dim + config.n_static_context
        n_output = config.n_state_with_flag
        self.lstm_layer, self.output_layer = self._setup_inner(
            n_input,
            n_output,
            config.n_hidden,
            config.n_layer,
            config.n_output_layer,
            config.variational,
        )

        self.parametric_bias_list = []
        for i in range(config.n_pb):
            name = "pb{}".format(i)
            param = Parameter(torch.zeros(config.n_pb_dim))
            self.register_parameter(name, param)
            self.parametric_bias_list.append(param)

    def loss(self, sample: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]) -> LossDict:
        episode_indices, state_sample, context_sample = sample
        self._sanity_check_context_sample(state_sample, context_sample)

        # create pb list
        assert self.parametric_bias_list is not None
        pb_list_extracted = [self.parametric_bias_list[i] for i in episode_indices]
        pb_stacked = torch.stack(pb_list_extracted)  # type: ignore

        # propagation
        state_sample_input, state_sample_output = state_sample[:, :-1], state_sample[:, 1:]
        pred, _ = self.forward(state_sample_input, pb_stacked, context_sample)
        return self._loss_inner(state_sample_output, pred)

    def forward(
        self,
        state_sample: torch.Tensor,
        pb_sample: torch.Tensor,
        context_sample: torch.Tensor,
        hidden: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:

        self._sanity_check_pb_sample(state_sample, pb_sample)
        self._sanity_check_context_sample(state_sample, context_sample)

        n_batch, n_seq_len, _ = state_sample.shape
        assert state_sample.ndim == 3
        assert pb_sample.ndim == 2
        assert len(state_sample) == len(pb_sample)

        # similar to normal LSTM ...
        pb_sample = pb_sample.unsqueeze(dim=1)
        pb_sample = pb_sample.expand(-1, n_seq_len, -1)
        context_auged_state_sample = torch.cat((state_sample, pb_sample), dim=2)

        preout, hidden = self.lstm_layer(context_auged_state_sample, hidden)
        out = self.output_layer(preout)
        assert hidden is not None
        return out, hidden

    def _sanity_check_pb_sample(self, state_sample: torch.Tensor, pb_sample: torch.Tensor) -> None:
        n_batch, n_seq_len, n_dim = state_sample.shape
        assert pb_sample.ndim == 2
        assert pb_sample.shape[0] == n_batch
        assert pb_sample.shape[1] == self.config.n_pb_dim



import argparse
from pathlib import Path

from mohou.default import auto_detect_autoencoder_type
from mohou.file import get_project_path
from mohou.model import VariationalAutoEncoder
from mohou.script_utils import (
    visualize_image_reconstruction,
    visualize_variational_autoencoder,
)
from mohou.setting import setting
from mohou.trainer import TrainCache
from mohou.types import EpisodeBundle

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-pn", type=str, default=setting.primary_project_name, help="project name")
    parser.add_argument("-pp", type=str, help="project path. preferred over pn.")
    parser.add_argument("-n", type=int, default=5, help="number of visualization")
    args = parser.parse_args()
    project_name: str = args.pn
    project_path_str: str = args.pp
    n_vis: int = args.n

    if project_path_str is None:
        project_path = get_project_path(project_name)
    else:
        project_path = Path(project_path_str)

    bundle = EpisodeBundle.load(project_path)

    ae_type = auto_detect_autoencoder_type(project_path)
    model = TrainCache.load(project_path, ae_type).best_model
    visualize_image_reconstruction(project_path, bundle, model, n_vis)

    if ae_type == VariationalAutoEncoder:
        visualize_variational_autoencoder(project_path)



import copy
import logging
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Type, Union

import numpy as np
import torch
from torch.utils.data import Dataset

from mohou.encoding_rule import EncodingRule
from mohou.types import AngleVector, ElementBase, EpisodeBundle, TerminateFlag
from mohou.utils import AnyT, assert_equal_with_message, assert_seq_list_list_compatible

logger = logging.getLogger(__name__)


@dataclass
class SequenceDatasetConfig:
    n_aug: int = 20
    cov_scale: float = 0.1

    def __new__(cls, *args, **kwargs):
        assert len(args) == 0, "please instantiate config only using kwargs"
        return super(SequenceDatasetConfig, cls).__new__(cls)

    def __post_init__(self):
        assert self.n_aug >= 0
        assert self.cov_scale < 1.0
        logger.info("sequence dataset config: {}".format(self))


@dataclass
class SequenceDataAugmentor:
    covmat: np.ndarray
    cov_scale: float

    @classmethod
    def from_seqs(cls, seq_list: List[np.ndarray], cov_scale: float, take_diff: bool = True):
        """construct augmentor.
        seq_list: sequence list used to compute covariance matrix
        take_diff: if True, covaraince is computed using state difference (e.g. {s[t] - s[t-1]}_{1:T})
        otherwise covariance is computed using s[t]_{1:T} sequence.
        """
        if take_diff:
            covmat = cls.compute_diff_covariance(seq_list)
        else:
            covmat = cls.compute_covariance(seq_list)
        return cls(covmat, cov_scale)

    @staticmethod
    def compute_covariance(state_seq_list: List[np.ndarray]) -> np.ndarray:
        state_diffs = np.vstack(state_seq_list)
        cov_mat = np.cov(state_diffs.T)
        return cov_mat

    @staticmethod
    def compute_diff_covariance(state_seq_list: List[np.ndarray]) -> np.ndarray:
        state_diff_list = []
        for state_seq in state_seq_list:
            diff = state_seq[1:, :] - state_seq[:-1, :]
            state_diff_list.append(diff)
        state_diffs = np.vstack(state_diff_list)
        cov_mat = np.cov(state_diffs.T)
        return cov_mat

    def apply(self, seq: np.ndarray) -> np.ndarray:
        """apply augmentation
        seq: 2dim array (n_seqlen, n_dim)
        """
        assert seq.ndim == 2
        assert seq.shape[1] == self.covmat.shape[0]

        covmat_scaled = self.covmat * self.cov_scale**2

        seq_copied = copy.deepcopy(seq)

        n_seqlen, n_dim = seq_copied.shape
        mean = np.zeros(n_dim)
        noises = np.random.multivariate_normal(mean, covmat_scaled, n_seqlen)
        return seq_copied + noises


@dataclass
class PaddingSequenceAligner:
    n_seqlen_target: int

    @classmethod
    def from_seqs(
        cls, seqs: Union[List[np.ndarray], List[List]], n_after_termination: int
    ) -> "PaddingSequenceAligner":
        n_seqlen_max = max([len(seq) for seq in seqs])
        n_seqlen_target = n_seqlen_max + n_after_termination
        return cls(n_seqlen_target)

    def apply(self, seq: AnyT) -> AnyT:  # TODO: specify type
        """get padded sequence based on seq.
        seq: np.ndarray or List
        if seq is [1, 2, 3, 4] and n_seqlen_target = 10, the output padded sequence looks like
        [1, 2, 3, 4, 4, 4, 4, 4, 4, 4]
        """
        seq = copy.deepcopy(seq)
        n_seqlen = len(seq)
        n_padding = self.n_seqlen_target - n_seqlen
        assert n_padding >= 0

        # TODO(HiroIshida) To remove type-ignores it's better to use singledispatch .
        # however we prefere simpilicy in this case.
        if isinstance(seq, list):
            padded_seq = seq + [copy.deepcopy(seq[-1]) for _ in range(n_padding)]
        elif isinstance(seq, np.ndarray):
            if n_padding == 0:
                padded_seq = seq  # type: ignore
            else:
                elem_last = seq[-1]
                padding_seq_shape = [n_padding] + [1 for _ in range(elem_last.ndim)]
                padding_seq = np.tile(elem_last, padding_seq_shape)
                padded_seq = np.concatenate((seq, padding_seq), axis=0)
        else:
            assert False

        assert len(padded_seq) == self.n_seqlen_target
        return padded_seq  # type: ignore


@dataclass
class AutoRegressiveDatasetConfig(SequenceDatasetConfig):
    n_dummy_after_termination: int = 20
    window_size: Optional[int] = None  # if None, all dataset will be just used.
    av_calibration_bias_std: float = 0.0


@dataclass
class AngleVectorCalibrationBiasRandomizer:
    type_bound_table: Dict[Type[ElementBase], slice]
    bias_std: float

    def apply(self, seq: np.ndarray) -> np.ndarray:
        """apply augmentor seq => seq unlike in SequenceDataAugmentor seq => List[seq]"""

        if AngleVector not in self.type_bound_table:
            return seq

        bound = self.type_bound_table[AngleVector]
        av_dim = bound.stop - bound.start
        calibration_error_vec = np.random.randn(av_dim) * self.bias_std

        seq_new = copy.deepcopy(seq)
        seq_new[:, bound] = seq_new[:, bound] + calibration_error_vec
        return seq_new


@dataclass
class AutoRegressiveDataset(Dataset):
    state_seq_list: List[np.ndarray]  # with flag info
    episode_index_list: List[int]
    static_context_list: List[np.ndarray]
    encoding_rule: EncodingRule

    def __post_init__(self):
        assert_equal_with_message(
            len(self.static_context_list), len(self.state_seq_list), "length of sequence"
        )

        # state sequence consists of n_seqlen x n_dim
        assert_equal_with_message(self.state_seq_list[0].ndim, 2, "dimension of state sequence")
        # context sequence consists of n_dim (because its static througout the context list)
        assert_equal_with_message(self.static_context_list[0].ndim, 1, "dimension of context")

    def __len__(self) -> int:
        return len(self.state_seq_list)

    def __getitem__(self, idx) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """return tuple of episode index of the state_seq, state_esq, and static context"""
        state_seq = torch.from_numpy(self.state_seq_list[idx]).float()
        context = torch.from_numpy(self.static_context_list[idx]).float()
        episode_index = torch.tensor(self.episode_index_list[idx], dtype=torch.int32)
        return episode_index, state_seq, context

    @classmethod
    def from_bundle(
        cls,
        bundle: EpisodeBundle,
        encoding_rule: EncodingRule,
        dataset_config: Optional[AutoRegressiveDatasetConfig] = None,
        static_context_list: Optional[List[np.ndarray]] = None,
    ) -> "AutoRegressiveDataset":

        last_key_of_rule = list(encoding_rule.keys())[-1]
        assert last_key_of_rule == TerminateFlag
        if dataset_config is None:
            dataset_config = AutoRegressiveDatasetConfig()

        state_seq_list = encoding_rule.apply_to_episode_bundle(bundle)

        # setting up static contect
        if static_context_list is None:  # create sequence of 0-dim vector
            static_context_list = [np.zeros((0)) for _ in range(len(state_seq_list))]
        assert_equal_with_message(
            len(static_context_list), len(state_seq_list), "length of sequence"
        )

        # augmentation
        randomizer1 = SequenceDataAugmentor.from_seqs(state_seq_list, dataset_config.cov_scale)
        randomizer2 = AngleVectorCalibrationBiasRandomizer(
            encoding_rule.type_bound_table, dataset_config.av_calibration_bias_std
        )

        episode_index_list_auged = []
        state_seq_list_auged = []
        static_context_list_auged = []

        for episode_index in range(len(state_seq_list)):
            seq = state_seq_list[episode_index]
            static_context = static_context_list[episode_index]

            # append the original
            episode_index_list_auged.append(episode_index)
            state_seq_list_auged.append(seq)
            static_context_list_auged.append(static_context)

            # append the augmented
            for _ in range(dataset_config.n_aug):
                seq_randomized = randomizer2.apply(randomizer1.apply(seq))

                episode_index_list_auged.append(episode_index)
                state_seq_list_auged.append(seq_randomized)
                static_context_list_auged.append(copy.deepcopy(static_context))

        if dataset_config.window_size is None:
            # make all sequence to the same length due to torch batch computation requirement
            aligner = PaddingSequenceAligner.from_seqs(
                state_seq_list_auged, dataset_config.n_dummy_after_termination
            )
            state_seq_list_auged_adjusted = [aligner.apply(seq) for seq in state_seq_list_auged]

            return cls(
                state_seq_list_auged_adjusted,
                episode_index_list_auged,
                static_context_list_auged,
                encoding_rule,
            )
        else:
            # split sequence by window size (Experimental feature!)
            window_state_seq_list = []
            window_episode_index_list = []
            window_static_context_list = []

            window_size = dataset_config.window_size

            for seq_idx in range(len(state_seq_list_auged)):
                state_seq = state_seq_list_auged[seq_idx]
                episode_idx = episode_index_list_auged[seq_idx]
                context = static_context_list_auged[seq_idx]

                n_window = len(state_seq) - window_size + 1
                for window_idx in range(n_window):
                    window_state_seq_list.append(state_seq[window_idx : window_idx + window_size])
                    window_episode_index_list.append(episode_idx)
                    window_static_context_list.append(context)
            return cls(
                window_state_seq_list,
                window_episode_index_list,
                window_static_context_list,
                encoding_rule,
            )


@dataclass
class MarkovControlSystemDataset(Dataset):
    """o_{t+1} = f(o_{t}, u_t{t})"""

    inp_ctrl_seq: np.ndarray
    inp_obs_seq: np.ndarray
    out_obs_seq: np.ndarray

    def __len__(self) -> int:
        return len(self.inp_ctrl_seq)

    def __getitem__(self, idx) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        inp_ctrl = torch.from_numpy(self.inp_ctrl_seq[idx]).float()
        inp_obs = torch.from_numpy(self.inp_obs_seq[idx]).float()
        out_obs = torch.from_numpy(self.out_obs_seq[idx]).float()
        return inp_ctrl, inp_obs, out_obs

    @classmethod
    def from_bundle(
        cls,
        bundle: EpisodeBundle,
        control_encoding_rule: EncodingRule,
        observation_encoding_rule: EncodingRule,
        config: Optional[SequenceDatasetConfig] = None,
        diff_as_control: bool = True,
    ) -> "MarkovControlSystemDataset":
        if config is None:
            config = SequenceDatasetConfig()

        ctrl_seq_list = control_encoding_rule.apply_to_episode_bundle(bundle)
        obs_seq_list = observation_encoding_rule.apply_to_episode_bundle(bundle)
        assert_seq_list_list_compatible([ctrl_seq_list, obs_seq_list])

        # augmentation
        ctrl_augmentor = SequenceDataAugmentor.from_seqs(
            ctrl_seq_list, config.cov_scale, take_diff=False
        )
        obs_augmentor = SequenceDataAugmentor.from_seqs(
            obs_seq_list, config.cov_scale, take_diff=True
        )

        ctrl_seq_list_auged = []
        obs_seq_list_auged = []
        for ctrl_seq, obs_seq in zip(ctrl_seq_list, obs_seq_list):
            ctrl_seq_list_auged.append(ctrl_seq)
            obs_seq_list_auged.append(obs_seq)
            for _ in range(config.n_aug):
                ctrl_seq_list_auged.append(ctrl_augmentor.apply(ctrl_seq))
                obs_seq_list_auged.append(obs_augmentor.apply(obs_seq))

        inp_ctrl_seq = []
        inp_obs_seq = []
        out_obs_seq = []
        for i in range(len(ctrl_seq_list_auged)):
            ctrl_seq = ctrl_seq_list_auged[i]
            obs_seq = obs_seq_list_auged[i]
            for j in range(len(ctrl_seq) - 1):
                if diff_as_control:
                    inp_ctrl_seq.append(ctrl_seq[j + 1] - ctrl_seq[j])
                else:
                    inp_ctrl_seq.append(ctrl_seq[j])
                inp_obs_seq.append(obs_seq[j])
                out_obs_seq.append(obs_seq[j + 1])

        return cls(np.array(inp_ctrl_seq), np.array(inp_obs_seq), np.array(out_obs_seq))



from dataclasses import dataclass
from typing import List, Optional, Tuple, Type

import torch
import torch.nn as nn

from mohou.model.common import LossDict, ModelBase, ModelConfigBase


def build_linear_layers(
    n_input: int, n_output: int, n_hidden: int, n_layer: int, activation: Optional[str]
) -> List[nn.Module]:

    if activation is not None:
        assert activation in ("relu", "sigmoid", "tanh")

    AT: Optional[Type[nn.Module]] = None
    if activation == "relu":
        AT = nn.ReLU
    elif activation == "sigmoid":
        AT = nn.Sigmoid
    elif activation == "tanh":
        AT = nn.Tanh

    layers: List[nn.Module] = []
    input_layer = nn.Linear(n_input, n_hidden)
    layers.append(input_layer)
    if AT is not None:
        layers.append(AT())

    for _ in range(n_layer):
        middle_layer = nn.Linear(n_hidden, n_hidden)
        layers.append(middle_layer)
        if AT is not None:
            layers.append(AT())

    output_layer = nn.Linear(n_hidden, n_output)
    layers.append(output_layer)
    return layers


@dataclass
class MarkoveModelConfig(ModelConfigBase):
    n_input: int
    n_output: int
    n_hidden: int = 200
    n_layer: int = 4
    activation: Optional[str] = None  # TODO(HiroIshida): consider replace it with enum

    def __post_init__(self):
        if self.activation is not None:
            assert self.activation in ("relu", "sigmoid", "tanh")


class ControlModel(ModelBase):
    layer: nn.Sequential

    def _setup_from_config(self, config: MarkoveModelConfig) -> None:
        config.n_input
        layers = build_linear_layers(
            n_input=config.n_input,
            n_output=config.n_output,
            n_hidden=config.n_hidden,
            n_layer=config.n_layer,
            activation=config.activation,
        )
        self.layer = nn.Sequential(*layers)

    def loss(self, sample: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]) -> LossDict:
        inp_ctrl_sample, inp_obs_sample, out_obs_sample = sample
        inp_sample = torch.concat((inp_ctrl_sample, inp_obs_sample), dim=1)
        out_obs = self.layer(inp_sample)
        loss = nn.MSELoss()(out_obs_sample, out_obs)
        return LossDict({"prediction": loss})
        return self.layer(sample)



import copy
from dataclasses import dataclass
from typing import List, Optional, Tuple, Type

import numpy as np
import torch
from torch.utils.data import Dataset

from mohou.dataset.sequence_dataset import (
    AutoRegressiveDatasetConfig,
    PaddingSequenceAligner,
    SequenceDataAugmentor,
)
from mohou.encoding_rule import EncodingRule
from mohou.types import EpisodeBundle, ImageBase, RGBImage
from mohou.utils import assert_equal_with_message, assert_seq_list_list_compatible


@dataclass
class ChimeraDataset(Dataset):
    image_type: Type[ImageBase]
    image_seqs: List[List[ImageBase]]
    vector_seqs: List[np.ndarray]

    def __post_init__(self):
        assert_equal_with_message(len(self.image_seqs), len(self.vector_seqs), "length of seq")

    def __len__(self) -> int:
        return len(self.image_seqs)

    def __getitem__(self, idx) -> Tuple[torch.Tensor, torch.Tensor]:
        image_seq = self.image_seqs[idx]
        image_seq_tensor = torch.stack([img.to_tensor() for img in image_seq], dim=0)

        vector_seq = self.vector_seqs[idx]
        vector_seq_tensor = torch.from_numpy(vector_seq).float()
        return image_seq_tensor, vector_seq_tensor

    @classmethod
    def from_bundle(
        cls,
        bundle: EpisodeBundle,
        encoding_rule: EncodingRule,
        config: Optional[AutoRegressiveDatasetConfig] = None,
    ) -> "ChimeraDataset":

        if config is None:
            config = AutoRegressiveDatasetConfig()

        # assume that encoding_rule does not have encoder for image
        for elem_type in encoding_rule.keys():
            assert not issubclass(elem_type, ImageBase)
        vector_seqs: List[np.ndarray] = encoding_rule.apply_to_episode_bundle(bundle)

        # TODO: currently assumes only RGB
        image_seqs: List[List[RGBImage]] = []
        for episode_data in bundle:
            tmp = episode_data.get_sequence_by_type(RGBImage)
            image_seqs.append(tmp.elem_list)

        # data augmentation
        augmentor = SequenceDataAugmentor.from_seqs(vector_seqs, config.cov_scale)
        vector_seqs_auged = []
        image_seqs_auged = []
        for image, seq in zip(image_seqs, vector_seqs):
            vector_seqs_auged.append(copy.deepcopy(seq))
            image_seqs_auged.append(image)

            for _ in range(config.n_aug):
                image_seqs_auged.append(image)
                vector_seqs_auged.append(augmentor.apply(seq))

        # align seq list
        n_after_termination = config.n_dummy_after_termination
        assert_seq_list_list_compatible([image_seqs_auged, vector_seqs_auged])
        aligner = PaddingSequenceAligner.from_seqs(image_seqs_auged, n_after_termination)
        image_seqs_aligned = [aligner.apply(seq) for seq in image_seqs_auged]
        vector_seqs_aligned = [aligner.apply(seq) for seq in vector_seqs_auged]
        return cls(RGBImage, image_seqs_aligned, vector_seqs_aligned)  # type: ignore



import argparse
from pathlib import Path
from typing import Optional, Type

from mohou.dataset import AutoEncoderDatasetConfig
from mohou.file import get_project_path
from mohou.model import AutoEncoder, AutoEncoderBase, VariationalAutoEncoder
from mohou.model.autoencoder import AutoEncoderConfig
from mohou.script_utils import create_default_logger, train_autoencoder
from mohou.setting import setting
from mohou.trainer import TrainConfig
from mohou.types import EpisodeBundle, ImageBase, RGBImage, get_element_type

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-pn", type=str, default=setting.primary_project_name, help="project name")
    parser.add_argument("-pp", type=str, help="project path. preferred over pn")
    parser.add_argument("-n", type=int, default=3000, help="iteration number")
    parser.add_argument("-aug", type=int, default=2, help="number of augmentation X")
    parser.add_argument("-latent", type=int, default=16, help="latent space dim")
    parser.add_argument("-image", type=str, default="RGBImage", help="image type")
    parser.add_argument("-bundle_postfix", type=str, default="", help="postfix for bundle")
    parser.add_argument(
        "-valid-ratio", type=float, default=0.1, help="split rate for validation dataset"
    )
    parser.add_argument("--vae", action="store_true", help="use vae")
    parser.add_argument("--warm", action="store_true", help="warm start")
    args = parser.parse_args()

    project_name: str = args.pn
    project_path_str: Optional[str] = args.pp
    n_epoch: int = args.n
    n_aug: int = args.aug
    n_bottleneck: int = args.latent
    valid_ratio: float = args.valid_ratio
    use_vae: bool = args.vae
    warm_start: bool = args.warm
    bundle_postfix: Optional[str] = args.bundle_postfix

    if project_path_str is None:
        project_path = get_project_path(project_name)
    else:
        project_path = Path(project_path_str)

    logger = create_default_logger(project_path, "autoencoder")

    if bundle_postfix == "":
        bundle_postfix = None
    bundle = EpisodeBundle.load(project_path, bundle_postfix)

    image_type: Type[ImageBase] = get_element_type(args.image)  # type: ignore
    n_pixel, _, _ = bundle.spec.type_shape_table[RGBImage]  # Assuming bundle contains rgb
    model_config = AutoEncoderConfig(image_type, n_bottleneck, n_pixel)
    dataset_config = AutoEncoderDatasetConfig(n_aug)
    train_config = TrainConfig(n_epoch=n_epoch, valid_data_ratio=valid_ratio)
    ae_type: Type[AutoEncoderBase] = VariationalAutoEncoder if use_vae else AutoEncoder  # type: ignore
    train_autoencoder(
        project_path,
        image_type,
        model_config,
        dataset_config,
        train_config,
        bundle=bundle,
        ae_type=ae_type,
        warm_start=warm_start,
    )



import copy
import logging
import operator
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import reduce
from typing import Any, Dict, Generic, List, Optional, TypeVar

import numpy as np
import torch
import torch.nn as nn

from mohou.utils import detect_device

logger = logging.getLogger(__name__)


class FloatLossDict(Dict[str, float]):
    def total(self) -> float:
        return reduce(operator.add, self.values())

    def __str__(self) -> str:
        def to_exp_notation(val: float) -> str:
            return np.format_float_scientific(val, precision=3, exp_digits=2)

        string = "total: {}".format(to_exp_notation(self.total()))
        for k, v in self.items():
            string += ", {}: {}".format(k, to_exp_notation(v))
        return string


def average_float_loss_dict(dicts: List[FloatLossDict]) -> FloatLossDict:
    dict_new = copy.deepcopy(dicts[0])
    for key in dict_new.keys():
        dict_new[key] = np.mean([d[key] for d in dicts]).item()
    return dict_new


class LossDict(Dict[str, torch.Tensor]):
    """A dictionary containing loss info.

    For example in VAE, loss is sum of reconstruction loss and regularization loss.
    Instead of returning the sum value directrly, returning LossDict in this case
    is beneficiall for debugging and
    visualization purposes.
    """

    def total(self) -> torch.Tensor:
        return reduce(operator.add, self.values())

    def to_float_lossdict(self) -> FloatLossDict:
        fld = FloatLossDict()
        for key in self.keys():
            fld[key] = self[key].detach().clone().cpu().item()
        return fld

    def __mul__(self, scale: float):
        return LossDict({key: val * scale for key, val in self.items()})

    def __rmul__(self, scale: float):
        return LossDict({key: val * scale for key, val in self.items()})


@dataclass
class ModelConfigBase:
    def to_dict(self) -> Dict:
        d: Dict[str, Any] = {}
        for key in self.__dataclass_fields__.keys():  # type: ignore
            val = self.__dict__[key]
            if isinstance(val, (int, float, str)):
                d[key] = val
            elif isinstance(val, type):
                d[key] = val.__name__
            elif val is None:
                d[key] = "none"
            elif isinstance(val, ModelConfigBase):
                d[key] = val.to_dict()
            else:
                message = str(val)
                message += (
                    " [note!] mohou > type {} is not supported yet. converted by str()".format(
                        type(val)
                    )
                )
                d[key] = message

        return d


ModelConfigT = TypeVar("ModelConfigT", bound=ModelConfigBase)


class ModelBase(nn.Module, Generic[ModelConfigT], ABC):
    config: ModelConfigT
    device: torch.device

    def __init__(self, config: ModelConfigT, device: Optional[torch.device] = None):
        super().__init__()
        self._setup_from_config(config)

        if device is None:
            device = detect_device()

        self.device = device
        self.config = config
        logger.info("model name: {}".format(self.__class__.__name__))
        logger.info("model config: {}".format(config))
        logger.info("model is initialized")

    def put_on_device(self, device: Optional[torch.device] = None):
        if device is not None:
            self.device = device
        self.to(self.device)

    @abstractmethod
    def _setup_from_config(self, config: ModelConfigT) -> None:
        pass

    @abstractmethod
    def loss(self, sample: Any) -> LossDict:
        pass

    @property
    def n_parameter(model):
        total_params = 0
        for name, parameter in model.named_parameters():
            if not parameter.requires_grad:
                continue
            params = parameter.numel()
            total_params += params
        return total_params


ModelT = TypeVar("ModelT", bound=ModelBase)



import functools
import pathlib
import queue
import subprocess
import types
from typing import Any, Callable, Iterator, List, Type, TypeVar, Union, cast

import numpy as np
import PIL
import torch
from torch.utils.data import Dataset, random_split

AnyT = TypeVar("AnyT", bound=Any)


class DataclassLightMixin:
    # sometimes, dataclass is too feature-rich and because of that
    # cause troubles. For example, dataclass is not straightly
    # incorpolated with abstract property.
    # To avoid this, this class provides minimalistic dataclass
    # implementation.
    def __init__(self, *args):
        assert len(args) == len(self.__annotations__)
        for key, arg in zip(self.__annotations__.keys(), args):
            setattr(self, key, arg)

    def __repr__(self) -> str:
        out = ""
        for key in self.__annotations__.keys():
            val = getattr(self, key)
            out += "{}: {} \n".format(key, repr(val))
        return out

    def __str__(self) -> str:
        return self.__repr__()


def abstract_attribute(obj: Callable[[Any], AnyT] = None) -> AnyT:
    class DummyAttribute:
        pass

    _obj = cast(Any, obj)
    if obj is None:
        _obj = DummyAttribute()
    _obj.__is_abstract_attribute__ = True
    return cast(AnyT, _obj)


def get_all_concrete_leaftypes(root: Type) -> List[Type]:
    concrete_types: List[Type] = []
    q = queue.Queue()  # type: ignore
    q.put(root)
    while not q.empty():
        t: Type = q.get()
        if len(t.__subclasses__()) == 0:
            concrete_types.append(t)

        for st in t.__subclasses__():
            q.put(st)
    return list(set(concrete_types))


def flatten_lists(list_list: List[List[AnyT]]) -> List[AnyT]:
    return functools.reduce(lambda xs, ys: xs + ys, list_list)


def get_bound_list(dims: List[int]) -> List[slice]:
    bound_list = []
    head = 0
    for dim in dims:
        bound_list.append(slice(head, head + dim))
        head += dim
    return bound_list


def splitting_slices(n_elem_list: List[int]) -> Iterator[slice]:
    head = 0
    for n_elem in n_elem_list:
        tail = head + n_elem
        yield slice(head, tail)
        head = tail


SequenceT = TypeVar("SequenceT")  # TODO(HiroIshida) bound?


def split_sequence(seq: SequenceT, n_elem_list: List[int]) -> Iterator[SequenceT]:
    for sl in splitting_slices(n_elem_list):
        yield seq[sl]  # type: ignore


def detect_device() -> torch.device:
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    return device


def split_with_ratio(dataset: Dataset, valid_ratio: float = 0.1):
    """split dataset into train and validation dataset with specified ratio"""

    n_total = len(dataset)  # type: ignore
    n_validate = int(valid_ratio * n_total)
    ds_train, ds_validate = random_split(dataset, [n_total - n_validate, n_validate])
    return ds_train, ds_validate


def assert_equal_with_message(given: AnyT, expected: Union[AnyT, List[Any]], elem_name: str):
    message = "{0}: given {1}, exepcted {2}".format(elem_name, given, expected)
    if isinstance(expected, list):
        assert given in expected, message
    else:
        assert given == expected, message


def assert_isinstance_with_message(given: Any, expected: Type):
    assert isinstance(given, expected), "{0}: given {1}, exepcted {2}".format(
        "not isinstance", given, expected
    )


def assert_seq_list_list_compatible(seq_llist: List[List[Any]]):
    if not __debug__:
        return
    seqlen_list_reference = [len(seq) for seq in seq_llist[0]]  # first seq_list as ref
    for seq_list in seq_llist:
        seqlen_list = [len(seq) for seq in seq_list]
        assert seqlen_list == seqlen_list_reference


def canvas_to_ndarray(fig, resize_pixel=None) -> np.ndarray:
    fig.canvas.draw()
    data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep="")
    data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    if resize_pixel is None:
        return data
    img = PIL.Image.fromarray(data)
    img_resized = img.resize(resize_pixel)
    data_resized = np.asarray(img_resized, dtype=np.uint8)
    return data_resized


def log_text_with_box(logger, text: str) -> None:
    box_width = max(40, len(text) * 3)
    inner_text = "=" * 10 + " " + text.capitalize() + " "
    inner_text += "=" * (box_width - len(inner_text))

    logger.info("=" * box_width)
    logger.info(inner_text)
    logger.info("=" * box_width)


def log_package_version_info(logger, module: types.ModuleType) -> None:
    def log_line_by_line(multiline_text: str):
        for line in multiline_text.splitlines():
            logger.info(line)

    init_file_name: str = module.__file__  # type: ignore
    is_site_package = "site-packages" in init_file_name
    log_text_with_box(logger, "version check")
    if is_site_package:
        logger.info("version: {}".format(module.__version__))  # type: ignore
    else:
        repo_path = pathlib.Path(init_file_name).parent.parent

        # git log
        command_git_log = "cd {}; git --no-pager log --oneline | head -20".format(str(repo_path))
        proc = subprocess.run(command_git_log, stdout=subprocess.PIPE, shell=True)
        git_log_stdout = proc.stdout.decode("utf8")
        logger.info(command_git_log)
        log_line_by_line(git_log_stdout)

        # git diff
        log_text_with_box(logger, "git diff check")
        command_git_diff = "cd {}; git --no-pager diff".format(str(repo_path))
        proc = subprocess.run(command_git_diff, stdout=subprocess.PIPE, shell=True)
        git_diff_stdout = proc.stdout.decode("utf8")
        logger.info(command_git_diff)
        log_line_by_line(git_diff_stdout)


def change_color_to_yellow(message: str) -> str:
    color = "\033[33m"
    reset = "\033[0m"
    return color + message + reset



import argparse
from pathlib import Path

from mohou.dataset import MarkovControlSystemDataset
from mohou.encoder import ImageEncoder, VectorIdenticalEncoder
from mohou.encoding_rule import EncodingRule
from mohou.file import get_project_path
from mohou.model import ControlModel, VariationalAutoEncoder
from mohou.model.markov import MarkoveModelConfig
from mohou.script_utils import create_default_logger
from mohou.setting import setting
from mohou.trainer import TrainCache, TrainConfig, train
from mohou.types import AngleVector, EpisodeBundle


def create_obs_rule(project_path: Path):
    tcache = TrainCache[VariationalAutoEncoder].load(project_path, VariationalAutoEncoder)
    model = tcache.best_model
    f = ImageEncoder.from_auto_encoder(model)
    obs_rule = EncodingRule.from_encoders([f])
    return obs_rule


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-pn", type=str, default=setting.primary_project_name, help="project name")
    parser.add_argument(
        "-full_path",
        type=str,
        help="project full path. If this arg is set, full path will be prefered over project name",
    )
    parser.add_argument("-n", type=int, default=3000, help="iteration number")

    args = parser.parse_args()
    project_name: str = args.pn
    n_epoch: int = args.n

    if args.full_path is None:
        project_path = get_project_path(project_name)
    else:
        project_path = Path(args.full_path)

    logger = create_default_logger(project_path, "control_model")

    bundle = EpisodeBundle.load(project_path)
    n_av_dim = bundle.spec.type_shape_table[AngleVector][0]
    f = VectorIdenticalEncoder.create(AngleVector, n_av_dim)
    ctrl_rule = EncodingRule.from_encoders([f])

    obs_rule = create_obs_rule(project_path)

    dataset = MarkovControlSystemDataset.from_bundle(
        bundle, ctrl_rule, obs_rule, diff_as_control=True
    )

    n_input = ctrl_rule.dimension + obs_rule.dimension
    n_output = obs_rule.dimension
    config = MarkoveModelConfig(n_input, n_output, activation="relu")
    tcache = TrainCache[ControlModel].from_model(ControlModel(config))
    tconfig = TrainConfig(n_epoch=n_epoch)
    train(project_path, tcache, dataset, config=tconfig)



import json
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from typing import Dict, Iterator, List, Mapping, Optional, Type, TypeVar, Union

import numpy as np
import torch

from mohou.encoder import EncoderBase, HasAModel
from mohou.types import (
    CompositeImageBase,
    ElementBase,
    ElementDict,
    EpisodeBundle,
    EpisodeData,
    PrimitiveElementBase,
)
from mohou.utils import assert_equal_with_message, get_bound_list

logger = logging.getLogger(__name__)


ScaleBalancerT = TypeVar("ScaleBalancerT", bound="ScaleBalancerBase")
ArrayT = TypeVar("ArrayT", bound=Union[torch.Tensor, np.ndarray])


class ScaleBalancerBase(ABC):
    @classmethod
    def get_json_file_path(cls, project_path: Path, create_dir: bool = False) -> Path:
        """get the json file path that will be used for dumping and loading
        Note that dumping / loading the scale balancer is still an experimental feature.
        """
        save_dir_path = project_path / "experimental"
        if create_dir:
            save_dir_path.mkdir(exist_ok=True)
        file_name = cls.__name__ + ".json"
        json_file_path = save_dir_path / file_name
        return json_file_path

    @abstractmethod
    def apply(self, arr: ArrayT) -> ArrayT:
        pass

    @abstractmethod
    def inverse_apply(self, arr: ArrayT) -> ArrayT:
        pass

    @abstractmethod
    def dump(self, project_path: Path) -> None:
        pass

    @classmethod
    @abstractmethod
    def load(cls: Type[ScaleBalancerT], project_path: Path) -> ScaleBalancerT:
        pass


@dataclass(frozen=True)
class IdenticalScaleBalancer(ScaleBalancerBase):
    def apply(self, arr: ArrayT) -> ArrayT:
        return arr

    def inverse_apply(self, arr: ArrayT) -> ArrayT:
        return arr

    def dump(self, project_path: Path) -> None:
        json_file_path = self.get_json_file_path(project_path, create_dir=True)
        with json_file_path.open(mode="w") as f:
            json.dump({}, f)

    @classmethod
    def load(cls, project_path: Path) -> "IdenticalScaleBalancer":
        logger.warning("NOTE: This feature may be dropped without any notification")
        json_file_path = cls.get_json_file_path(project_path)
        assert json_file_path.exists()
        with json_file_path.open(mode="r") as f:
            kwargs = json.load(f)
        # because actually this class does not have any members
        assert len(kwargs) == 0
        return cls()

    def __eq__(self, other: object) -> bool:
        keys = self.__dataclass_fields__.keys()  # type: ignore # mypy's bag
        assert set(keys) == set()
        return isinstance(other, IdenticalScaleBalancer)


@dataclass(frozen=True)
class CovarianceBasedScaleBalancer(ScaleBalancerBase):
    dims: List[int]
    means: List[np.ndarray]
    scaled_stds: List[float]

    def __eq__(self, other: object) -> bool:
        keys = self.__dataclass_fields__.keys()  # type: ignore # mypy's bag
        assert set(keys) == {"dims", "means", "scaled_stds"}

        if not isinstance(other, CovarianceBasedScaleBalancer):
            return NotImplemented
        assert type(self) == type(other)

        if self.dims != other.dims:
            return False
        if not np.allclose(np.hstack(self.means), np.hstack(other.means), atol=1e-6):
            return False
        if self.scaled_stds != other.scaled_stds:
            return False

        return True

    def dump(self, project_path: Path):
        json_file_path = self.get_json_file_path(project_path, create_dir=True)
        dic = {
            "dims": self.dims,
            "means": [mean.tolist() for mean in self.means],
            "scaled_stds": self.scaled_stds,
        }
        with json_file_path.open(mode="w") as f:
            json.dump(dic, f)

    @classmethod
    def load(cls, project_path: Path) -> "CovarianceBasedScaleBalancer":
        logger.warning("NOTE: This feature may be dropped without any notification")
        json_file_path = cls.get_json_file_path(project_path)
        assert json_file_path.exists()
        with json_file_path.open(mode="r") as f:
            kwargs = json.load(f)
        means = kwargs["means"]
        kwargs["means"] = [np.array(mean) for mean in means]
        return cls(**kwargs)

    def __post_init__(self):
        for i, dim in enumerate(self.dims):
            assert_equal_with_message(len(self.means), len(self.dims), "len")
            assert_equal_with_message(len(self.scaled_stds), len(self.dims), "len")
            assert_equal_with_message(self.means[i].shape, (dim,), "mean shape of {}".format(i))

    @staticmethod
    def get_bound_list(dims: List[int]) -> List[slice]:
        bound_list = []
        head = 0
        for dim in dims:
            bound_list.append(slice(head, head + dim))
            head += dim
        return bound_list

    @staticmethod
    def is_binary_sequence(partial_feature_seq: np.ndarray):
        return len(set(partial_feature_seq.flatten().tolist())) == 2

    @classmethod
    def from_feature_seqs(cls, feature_seq: np.ndarray, dims: List[int]):
        assert_equal_with_message(feature_seq.ndim, 2, "feature_seq.ndim")
        means = []
        max_stds = []
        for rang in cls.get_bound_list(dims):
            feature_seq_partial = feature_seq[:, rang]
            dim = feature_seq_partial.shape[1]
            if cls.is_binary_sequence(feature_seq_partial):
                # because it's strange to compute covariance for binary sequence
                assert dim == 1, "this restriction maybe removed"
                minn = np.min(feature_seq_partial)
                maxx = np.max(feature_seq_partial)
                cov = np.diag(np.ones(dim))
                mean = np.array([0.5 * (minn + maxx)])
            else:
                mean = np.mean(feature_seq_partial, axis=0)
                cov = np.cov(feature_seq_partial.T)
                if cov.ndim == 0:  # unfortunately, np.cov return 0 dim array instead of 1x1
                    cov = np.expand_dims(cov, axis=0)
                    cov = np.array([[cov.item()]])

            eig_values, _ = np.linalg.eig(cov)
            max_std = np.sqrt(max(eig_values))

            # Note that np.cov(pts.T) outputs some non 0 value even if the input points
            # are all at a single point (degenerated). So, max_eig > 0.0 cannot check
            # the data degeneration. Thus...
            assert max_std > 1e-7

            means.append(mean)
            max_stds.append(max_std)

        scaled_stds = list(np.array(max_stds) / max(max_stds))
        return cls(dims, means, scaled_stds)

    def _apply(self, arr: ArrayT, inverse: bool) -> ArrayT:

        assert arr.ndim in [1, 2]

        dim = len(arr) if arr.ndim == 1 else arr.shape[1]
        assert_equal_with_message(dim, sum(self.dims), "vector total dim")

        mean = np.hstack(self.means)
        std_list = [np.ones(dim) * std for dim, std, in zip(self.dims, self.scaled_stds)]
        std = np.hstack(std_list)

        if isinstance(arr, torch.Tensor):
            mean = torch.from_numpy(mean).float().to(arr.device)  # type: ignore
            std = torch.from_numpy(std).float().to(arr.device)  # type: ignore

        if inverse:
            return (arr * std) + mean  # type: ignore
        else:
            return (arr - mean) / std  # type: ignore

    def apply(self, arr: ArrayT) -> ArrayT:
        return self._apply(arr, False)

    def inverse_apply(self, arr: ArrayT) -> ArrayT:
        return self._apply(arr, True)


class EncodingRuleBase(Mapping[Type[ElementBase], EncoderBase]):
    @abstractmethod
    def apply(self, elem_dict: ElementDict) -> np.ndarray:
        pass

    @abstractmethod
    def inverse_apply(self, vector_processed: np.ndarray) -> ElementDict:
        pass

    @property
    @abstractmethod
    def dimension_list(self) -> List[int]:
        """get dimension list for each encoders output"""

    @property
    def type_bound_table(self) -> Dict[Type[ElementBase], slice]:
        bounds = get_bound_list(self.dimension_list)

        table = {}
        for encoder, bound in zip(self.values(), bounds):
            table[encoder.elem_type] = bound
        return table

    def apply_to_episode_data(self, episode_data: EpisodeData) -> np.ndarray:
        vec_list = [self.apply(episode_data[i]) for i in range(len(episode_data))]
        return np.array(vec_list)

    def apply_to_episode_bundle(self, bundle: EpisodeBundle) -> List[np.ndarray]:
        def elem_types_to_primitive_elem_set(elem_type_list: List[Type[ElementBase]]):
            primitve_elem_type_list = []
            for elem_type in elem_type_list:
                if issubclass(elem_type, PrimitiveElementBase):
                    primitve_elem_type_list.append(elem_type)
                elif issubclass(elem_type, CompositeImageBase):
                    primitve_elem_type_list.extend(elem_type.image_types)
            return set(primitve_elem_type_list)

        bundle_elem_types = elem_types_to_primitive_elem_set(list(bundle.types()))
        required_elem_types = elem_types_to_primitive_elem_set(list(self.keys()))
        assert required_elem_types <= bundle_elem_types

        vector_seq_list = [self.apply_to_episode_data(data) for data in bundle]

        assert vector_seq_list[0].ndim == 2
        return vector_seq_list

    def get_device(self) -> Optional[torch.device]:
        device_list = []
        for encoder in self.values():
            if isinstance(encoder, HasAModel):
                device_list.append(encoder.get_device())
        if len(device_list) == 0:
            return None
        device_set = set(device_list)
        assert len(device_set) == 1, "Do not mix more than 1 devices"
        return device_set.pop()

    def set_device(self, device: torch.device) -> None:
        for encoder in self.values():
            if isinstance(encoder, HasAModel):
                encoder.set_device(device)


class EncodingRule(Dict[Type[ElementBase], EncoderBase], EncodingRuleBase):
    scale_balancer: ScaleBalancerBase

    def apply(self, elem_dict: ElementDict) -> np.ndarray:
        vector_list = []
        for elem_type, encoder in self.items():
            vector = encoder.forward(elem_dict[elem_type])
            vector_list.append(vector)
        return self.scale_balancer.apply(np.hstack(vector_list))

    def inverse_apply(self, vector_processed: np.ndarray) -> ElementDict:
        def split_vector(vector: np.ndarray, size_list: List[int]):
            head = 0
            vector_list = []
            for i, size in enumerate(size_list):
                tail = head + size
                vector_list.append(vector[head:tail])
                head = tail
            return vector_list

        vector = self.scale_balancer.inverse_apply(vector_processed)
        size_list = [encoder.output_size for elem_type, encoder in self.items()]
        vector_list = split_vector(vector, size_list)

        elem_dict = ElementDict([])
        for vec, (elem_type, encoder) in zip(vector_list, self.items()):
            elem_dict[elem_type] = encoder.backward(vec)
        return elem_dict

    @property
    def dimension(self) -> int:
        return sum(encoder.output_size for encoder in self.values())

    @property
    def dimension_list(self) -> List[int]:
        return [encoder.output_size for encoder in self.values()]

    @property
    def encode_order(self) -> List[Type[ElementBase]]:
        return list(self.keys())

    def __str__(self) -> str:
        string = "total dim: {}".format(self.dimension)
        for elem_type, encoder in self.items():
            string += "\n{0}: {1}".format(elem_type.__name__, encoder.output_size)
        return string

    @classmethod
    def from_encoders(
        cls,
        encoder_list: List[EncoderBase],
        bundle: Optional[EpisodeBundle] = None,
        scale_balancer: Optional[ScaleBalancerBase] = None,
    ) -> "EncodingRule":
        """Create EncodingRule from encoder_list
        Args:
            encoder_list: list of encoder. Order of the list is important and preserved.
            bundle: If set, ScaleBalancer will created using bundle
            scale_balancer: use this scale balancer if set.

        bundle != None and scale_balancer != None is never accepted
        """
        # NOTE: currently we can load cached balancer. But loading cached entire encoder is,
        # of course the future direction. However, the difficulty mainly lies in the serializing
        # ImageEncoder which contains lambda functions.

        rule: EncodingRule = cls()
        for encoder in encoder_list:
            rule[encoder.elem_type] = encoder
        rule.scale_balancer = IdenticalScaleBalancer()  # tmp

        if scale_balancer is not None:
            assert not bundle
            rule.scale_balancer = scale_balancer

        if bundle is not None:
            assert not scale_balancer
            dims = [encoder.output_size for encoder in rule.values()]
            # compute normalizer and set to encoder
            vector_seqs = rule.apply_to_episode_bundle(bundle)
            vector_seq_concated = np.concatenate(vector_seqs, axis=0)
            rule.scale_balancer = CovarianceBasedScaleBalancer.from_feature_seqs(
                vector_seq_concated, dims
            )
        return rule


@dataclass
class CompositeEncodingRule(EncodingRuleBase):
    rules: List[EncodingRule]

    def __post_init__(self):
        key_set_list = [set(rule.keys()) for rule in self.rules]
        set_entire = set.union(*key_set_list)
        no_intersection = len(set_entire) == sum([len(rule.keys()) for rule in self.rules])
        assert no_intersection

    def apply(self, edict: ElementDict) -> np.ndarray:
        vec = np.hstack([rule.apply(edict) for rule in self.rules])
        return vec

    def inverse_apply(self, vector: np.ndarray) -> ElementDict:
        head = 0
        elems = []
        for rule in self.rules:
            tail = head + rule.dimension
            edict = rule.inverse_apply(vector[head:tail])
            elems.extend(list(edict.values()))
            head = tail
        edict_merged = ElementDict(elems)
        return edict_merged

    @property
    def dimension_list(self) -> List[int]:
        dims = []
        for rule in self.rules:
            for encoder in rule.values():
                dims.append(encoder.output_size)
        return dims

    def __getitem__(self, key: Type[ElementBase]) -> EncoderBase:
        for rule in self.rules:
            if key in rule:
                return rule[key]
        raise KeyError

    def __iter__(self) -> Iterator[Type[ElementBase]]:
        return chain(*[rule.__iter__() for rule in self.rules])

    def __len__(self) -> int:
        return sum([len(rule) for rule in self.rules])



import argparse
from pathlib import Path
from typing import Optional

from mohou.dataset import AutoRegressiveDatasetConfig
from mohou.default import create_default_encoding_rule
from mohou.file import get_project_path
from mohou.model.lstm import PBLSTM, PBLSTMConfig
from mohou.script_utils import create_default_logger, train_lstm
from mohou.setting import setting
from mohou.trainer import TrainConfig
from mohou.types import EpisodeBundle

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-pn", type=str, default=setting.primary_project_name, help="project name")
    parser.add_argument("-pp", type=str, help="project path. preferred over pn.")
    parser.add_argument("-n", type=int, default=3000, help="iteration number")
    parser.add_argument("-aug", type=int, default=2, help="number of augmentation X")
    parser.add_argument("-hidden", type=int, default=200, help="number of hidden state of lstm")
    parser.add_argument("-layer", type=int, default=4, help="number of layers of lstm")
    parser.add_argument("-cov-scale", type=float, default=0.1, help="covariance scale in aug")
    parser.add_argument("-pb", type=int, default=2, help="parametric bias dimension")
    parser.add_argument(
        "-valid-ratio", type=float, default=0.1, help="split rate for validation dataset"
    )
    parser.add_argument("--warm", action="store_true", help="warm start")
    args = parser.parse_args()

    project_name: str = args.pn
    project_path_str: Optional[str] = args.pp
    n_epoch: int = args.n
    n_aug: int = args.aug
    n_hidden: int = args.hidden
    n_layer: int = args.layer
    n_pb_dim: int = args.pb
    cov_scale: float = args.cov_scale
    valid_ratio: float = args.valid_ratio
    warm_start: bool = args.warm

    if project_path_str is None:
        project_path = get_project_path(project_name)
    else:
        project_path = Path(project_path_str)

    logger = create_default_logger(project_path, "pblstm")  # noqa

    bundle = EpisodeBundle.load(project_path)

    encoding_rule = create_default_encoding_rule(project_path)
    model_config = PBLSTMConfig(
        encoding_rule.dimension,
        n_hidden=n_hidden,
        n_layer=n_layer,
        n_pb_dim=n_pb_dim,
        n_pb=len(bundle),
    )
    dataset_config = AutoRegressiveDatasetConfig(n_aug=n_aug, cov_scale=cov_scale)
    train_config = TrainConfig(n_epoch=n_epoch, valid_data_ratio=valid_ratio)

    train_lstm(
        project_path,
        encoding_rule,
        model_config,
        dataset_config,
        train_config,
        model_type=PBLSTM,
        warm_start=warm_start,
    )



# from https://github.com/salesforce/awd-lstm-lm/blob/master/LICENSE
# BSD 3-Clause License
# Copyright (c) 2017,
# All rights reserved.

# from https://github.com/mourga/variational-lstm/blob/master/LICENSE
# MIT License
# Copyright (c) 2021 Katerina Margatina

import torch
import torch.nn as nn
from torch.autograd import Variable
from torch.nn.functional import dropout
from torch.nn.parameter import Parameter


class LockedDropout(nn.Module):
    """
    This function applies dropout to the input tensor x.
    The shape of the tensor x in our implementation is (batch_size, seq_len, feature_size)
    (contrary to Merity's AWD that uses (seq_len, batch_size, feature_size)).
    So, we sample a mask from the 'feature_size' dimension,
    but a different mask for each 'batch_size' dimension,
    and expand it in the 'sequence_length' dimension so that
    we apply the SAME mask FOR EACH TIMESTEP of the RNN (= 'seq_len' dim.).

    Code from https://github.com/salesforce/awd-lstm-lm
    paper: https://arxiv.org/pdf/1708.02182.pdf (see Section 4.2)
    """

    def __init__(self):
        super().__init__()

    def forward(self, x, dropout=0.5):
        if not self.training or not dropout:
            return x
        batch_size, seq_length, feat_size = x.size()
        m = x.data.new(batch_size, 1, feat_size).bernoulli_(1 - dropout)
        mask = Variable(m, requires_grad=False) / (1 - dropout)
        mask = mask.expand_as(x)
        return mask * x


class WeightDrop(torch.nn.Module):
    def __init__(self, module, weights, dropout=0, variational=True):
        """
        Dropout class that is paired with a torch module to make sure that the SAME mask
        will be sampled and applied to ALL timesteps.
        :param module: nn. module (e.g. nn.Linear, nn.LSTM)
        :param weights: which weights to apply dropout (names of weights of module)
        :param dropout: dropout to be applied
        :param variational: if True applies Variational Dropout, if False applies DropConnect (different masks!!!)

        Code from https://github.com/salesforce/awd-lstm-lm
        paper: https://arxiv.org/pdf/1708.02182.pdf
        for a difference between dropout and drop connect see https://stats.stackexchange.com/questions/201569/what-is-the-difference-between-dropout-and-drop-connect?fbclid=IwAR26B772sQXXBBpNSI_cu41zIEFu0gVkKgn34E1Wbbb4EbQdkdvmMOgro4s
        """
        super(WeightDrop, self).__init__()
        self.module = module
        self.weights = weights
        self.dropout = dropout
        self.variational = variational
        self._setup()

    def widget_demagnetizer_y2k_edition(*args, **kwargs):
        """
        Smerity code I don't understand.
        """
        # We need to replace flatten_parameters with a nothing function
        # It must be a function rather than a lambda as otherwise pickling explodes
        # We can't write boring code though, so ... WIDGET DEMAGNETIZER Y2K EDITION!
        # (╯°□°）╯︵ ┻━┻
        return

    def _setup(self):
        """
        This function renames each 'weight name' to 'weight name' + '_raw'
        (e.g. weight_hh_l0 -> weight_hh_l0_raw)
        :return:
        """
        # Terrible temporary solution to an issue regarding compacting weights re: CUDNN RNN
        if issubclass(type(self.module), torch.nn.RNNBase):
            self.module.flatten_parameters = self.widget_demagnetizer_y2k_edition

        for name_w in self.weights:
            print("Applying weight drop of {} to {}".format(self.dropout, name_w))
            w = getattr(self.module, name_w)
            del self.module._parameters[name_w]
            self.module.register_parameter(name_w + "_raw", Parameter(w.data))

    def _setweights(self):
        """
        This function samples & applies a dropout mask to the weights of the recurrent layers.
        Specifically, for an LSTM, each gate has
        - a W matrix ('weight_ih') that is multiplied with the input (x_t)
        - a U matrix ('weight_hh') that is multiplied with the previous hidden state (h_t-1)
        We sample a mask (either with Variational Dropout or with DropConnect) and apply it to
        the matrices U and/or W.
        The matrices to be dropped-out are in self.weights.
        A 'weight_hh' matrix is of shape (4*nhidden, nhidden)
        while a 'weight_ih' matrix is of shape (4*nhidden, ninput).

        **** Variational Dropout ****
        With this method, we sample a mask from the tensor (4*nhidden, 1) PER ROW
        and expand it to the full matrix.

        **** DropConnect ****
        With this method, we sample a mask from the tensor (4*nhidden, nhidden) directly
        which means that we apply dropout PER ELEMENT/NEURON.
        :return:
        """
        for name_w in self.weights:
            raw_w = getattr(self.module, name_w + "_raw")
            w = None

            mask = torch.autograd.Variable(torch.ones(raw_w.size(0), 1))
            if raw_w.is_cuda:
                mask = mask.cuda()
            mask = dropout(mask, p=self.dropout, training=True)  # type: ignore
            w = mask.expand_as(raw_w) * raw_w
            setattr(self.module, name_w, w)

    def forward(self, *args):
        self._setweights()
        return self.module.forward(*args)


class VariationalLSTM(nn.Module):
    def __init__(
        self,
        ninput,
        nhidden,
        nlayers=1,
        bidirectional=False,
        dropouti=0.0,
        dropoutw=0.0,
        dropouto=0.0,
        dropout=0.0,
        pack=True,
        last=False,
    ):
        """
        A simple RNN Encoder, which produces a fixed vector representation
        for a variable length sequence of feature vectors, using the output
        at the last timestep of the RNN.
        We use batch_first=True for our implementation.
        Tensors are are shape (batch_size, sequence_length, feature_size).
        Args:
            input_size (int): the size of the input features
            rnn_size (int):
            num_layers (int):
            bidirectional (bool):
            dropout (float):
        """
        super(VariationalLSTM, self).__init__()

        self.pack = pack
        self.last = last

        self.lockdrop = LockedDropout()

        if not isinstance(nhidden, list):
            nhidden = [nhidden for _ in range(nlayers)]

        self.ninp = ninput
        self.nhid = nhidden
        self.nlayers = nlayers
        self.dropouti = dropouti  # rnn input dropout
        self.dropoutw = dropoutw  # rnn recurrent dropout
        self.dropouto = dropouto  # rnn output dropout
        if dropout == 0.0 and dropouto != 0.0:
            self.dropout = self.dropouto  # rnn output dropout (of the last RNN layer)

        lstms = [
            nn.LSTM(
                input_size=ninput if i == 0 else nhidden[i - 1],
                hidden_size=nhidden[i],
                num_layers=1,
                dropout=0,
                batch_first=True,
            )
            for i in range(nlayers)
        ]

        # Dropout to recurrent layers (matrices weight_hh AND weight_ih of each layer of the RNN)
        dropped_lstms = [
            WeightDrop(rnn, ["weight_hh_l0", "weight_ih_l0"], dropout=dropoutw) for rnn in lstms
        ]

        self.lstms = nn.ModuleList(dropped_lstms)
        # self.init_weights()

    def reorder_hidden(self, hidden, order):
        """
        :param hidden:
        :param order:
        :return:
        """
        if isinstance(hidden, tuple):
            hidden = hidden[0][:, order, :], hidden[1][:, order, :]
        else:
            hidden = hidden[:, order, :]

        return hidden

    def init_hidden(self, bsz):
        """
        Initialise the hidden and cell state (h0, c0) for the first timestep (t=0).
        Both h0, c0 are of shape (num_layers * num_directions, batch_size, hidden_size)
        :param bsz: batch size
        :return:
        """
        weight = next(self.parameters()).data
        return [
            (
                weight.new(1, bsz, self.nhid[i]).zero_(),
                weight.new(1, bsz, self.nhid[i]).zero_(),
            )
            for i in range(self.nlayers)
        ]

    def forward(self, x, hidden=None, lengths=None, return_h=False):
        """

        :param x: tensor of shape (batch_size, seq_len, embedding_size)
        :param hidden: tuple (h0, c0), each of shape (num_layers * num_directions, batch_size, hidden_size)
        :param lengths: tensor (size 1 with true lengths)
        :return:
        """
        batch_size, seq_length, feat_size = x.size()

        # Dropout to inputs of the RNN (dropouti)
        emb = self.lockdrop(x, self.dropouti)

        if hidden is None:
            hidden = self.init_hidden(batch_size)

        raw_output = emb  # input to the first layer
        new_hidden = []
        raw_outputs = []
        outputs = []

        # for each layer of the RNN
        for i, rnn in enumerate(self.lstms):
            # calculate hidden states and output from the l RNN layer
            raw_output, new_h = rnn(raw_output, hidden[i])
            # save them in lists
            new_hidden.append(new_h)
            raw_outputs.append(raw_output)
            if i != self.nlayers - 1:
                # apply dropout to the output of the l-th RNN layer (dropouto)
                raw_output = self.lockdrop(raw_output, self.dropouto)
                # save 'dropped-out outputs' in a list
                outputs.append(raw_output)
        hidden = new_hidden

        # Dropout to the output of the last RNN layer (dropout)
        output = self.lockdrop(raw_output, self.dropout)
        outputs.append(output)

        # result = output.view(output.size(0) * output.size(1), output.size(2))
        result = output
        # result: output of the last RNN layer
        # hidden: hidden state of the last RNN layer
        # raw_outputs: outputs of all RNN layers without dropout
        # outputs: dropped-out outputs of all RNN layers
        if return_h:
            return result, hidden, raw_outputs, outputs
        return result, hidden



import datetime
import logging
import os
import pickle
import re
import shutil
from pathlib import Path
from typing import Any, List, Optional, Type, TypeVar, Union

from mohou.setting import setting

logger = logging.getLogger(__name__)


def get_root_path() -> Path:
    path = setting.root_path
    path.mkdir(exist_ok=True)
    return path


def create_project_dir(project_name: str) -> None:
    root_path = get_root_path()
    project_path = root_path / project_name
    project_path.mkdir(exist_ok=True)


def get_project_path(project_name: Optional[str] = None) -> Path:
    root_path = get_root_path()
    if project_name is None:
        assert setting.primary_project_name is not None
        project_name = setting.primary_project_name
    project_dir_path = root_path / project_name
    return project_dir_path


def get_subproject_path(project_name: Optional[str], subpath: Union[str, Path]) -> Path:

    if isinstance(subpath, str):
        subpath = Path(subpath)

    path = get_project_path(project_name)
    path = path / subpath
    path.mkdir(parents=True, exist_ok=True)
    return path


def remove_project(project_name: str) -> None:
    path = get_project_path(project_name)
    shutil.rmtree(str(path))


def resolve_file_path(
    obj_type: Type,
    project_name: Optional[str] = None,
    postfix: Optional[str] = None,
    subpath: Optional[Path] = None,
) -> Path:

    dir_path = get_project_path(project_name)

    if subpath is not None:
        dir_path = dir_path / subpath

    dir_path.mkdir(parents=True, exist_ok=True)

    if postfix is None:
        file_path = dir_path / (obj_type.__name__ + ".pkl")
    else:
        file_path = dir_path / (obj_type.__name__ + "-" + postfix + ".pkl")
    return file_path


DataT = TypeVar("DataT")


def load_object(
    obj_type: Type[DataT],
    project_name: Optional[str] = None,
    postfix: Optional[str] = None,
    subpath: Optional[Path] = None,
) -> DataT:
    """load single pickle object"""

    file_path = resolve_file_path(obj_type, project_name, postfix=postfix, subpath=subpath)
    time_stamp = os.path.getmtime(str(file_path))
    modified_time = datetime.datetime.fromtimestamp(time_stamp)

    logger.info(
        "load pickle from {} (modification time is {})".format(str(file_path), modified_time)
    )

    with file_path.open(mode="rb") as f:
        obj = pickle.load(f)

    return obj


def load_objects(
    obj_type: Type[DataT],
    project_name: Optional[str] = None,
    postfix: Optional[str] = None,
    subpath: Optional[Path] = None,
) -> List[DataT]:
    """load multiple pickle objects (or could be single object)
    If postfix is specified, all the objects filename of which starts with the postfix will be loaded.

    For example, if obj_type = SomeClass and postfix is "12345", file names
    path_to_project/Someclass-12345.pkl
    path_to_project/Someclass-12345-jfoiwfjoi.pkl
    will be loaded but
    path_to_project/Someclass-54321.pkl
    will not be loaded.
    """

    file_path = resolve_file_path(obj_type, project_name, postfix=postfix, subpath=subpath)
    file_name_common = str(file_path)
    base, file_name_common_local = os.path.split(file_name_common)
    _, ext = os.path.splitext(file_name_common_local)

    obj_list: List[DataT] = []
    for file_name in os.listdir(base):
        file_name_common_local_noext, _ = os.path.splitext(file_name_common_local)
        result = re.match(r"{}*.".format(file_name_common_local_noext), file_name)
        if result is not None:
            whole_name = os.path.join(base, file_name)
            with open(whole_name, "rb") as f:
                obj = pickle.load(f)
            obj_list.append(obj)
    if len(obj_list) == 0:
        raise FileNotFoundError

    return obj_list


def dump_object(
    obj: Any,
    project_name: Optional[str] = None,
    postfix: Optional[str] = None,
    subpath: Optional[Path] = None,
) -> None:

    file_path = resolve_file_path(obj.__class__, project_name, postfix=postfix, subpath=subpath)
    logger.info("dump pickle to {}".format(file_path))

    # Not using with statement to use custom exception handling
    file_name = str(file_path)
    f = open(file_name, "wb")
    try:
        pickle.dump(obj, f)
        f.close()
    except KeyboardInterrupt:
        f.close()
        logger.info("got keyboard interuppt. but let me dump the object...")
        with open(file_name, "wb") as f:
            pickle.dump(obj, f)
    except Exception as e:
        f.close()
        raise e



import subprocess
import tempfile
from pathlib import Path

import gdown


def get_asset_cache_path() -> Path:
    return Path("~/.mohou/asset").expanduser()


def get_panda_urdf_path() -> Path:
    """get franka_panda's urdf path
    If not file exists, model file will be downloaded.
    NOTE that, the urdf model is modified.
    """
    asset_cache_path = get_asset_cache_path()
    urdf_base_path = asset_cache_path / "franka_panda"
    urdf_file_path = urdf_base_path / "panda.urdf"

    if not urdf_file_path.exists():
        with tempfile.TemporaryDirectory() as td:
            urdf_base_path.mkdir(exist_ok=True, parents=True)

            temp_tarball_path = Path(td) / "tmp.tar"
            url = "https://drive.google.com/uc?id=1gZwXO1jDISZOVIFkC6exFAAk8JLeKco5"
            gdown.download(url, str(temp_tarball_path))
            assert temp_tarball_path.exists()
            subprocess.run(
                "cd {} && tar xf {}".format(urdf_base_path, temp_tarball_path), shell=True
            )
    assert urdf_file_path.exists()
    return urdf_file_path



from typing import Callable, Optional

import albumentations as al
import numpy as np

RandImageFunc = Optional[Callable[[np.ndarray], np.ndarray]]
_f_randomize_rgb_image: RandImageFunc = None
_f_randomize_gray_image: RandImageFunc = None
_f_randomize_depth_image: RandImageFunc = None


def configure_rgb_image_randomizer(rgb_shift_limit=40):
    def randomize_rgb_image(image_arr: np.ndarray):
        aug_guass = al.GaussNoise(p=1)
        sl = rgb_shift_limit
        aug_rgbshit = al.RGBShift(r_shift_limit=sl, g_shift_limit=sl, b_shift_limit=sl)
        aug_composed = al.Compose([aug_guass, aug_rgbshit])
        return aug_composed(image=image_arr)["image"]

    global _f_randomize_rgb_image
    _f_randomize_rgb_image = randomize_rgb_image


def configure_gray_image_randomizer():
    def randomize_gray_image(image_arr: np.ndarray):
        aug_guass = al.GaussNoise(p=1)
        aug_composed = al.Compose([aug_guass])
        return aug_composed(image=image_arr)["image"]

    global _f_randomize_gray_image
    _f_randomize_gray_image = randomize_gray_image


def configure_depth_image_randomizer(depth_shift_std=0.3, depth_noise_std=0.01):
    def randomize_depth_image(image_arr: np.ndarray):
        shift = np.random.randn() * depth_shift_std
        noise = np.random.randn(*image_arr.shape) * depth_noise_std
        image_arr_out = image_arr + noise + shift
        return image_arr_out

    global _f_randomize_depth_image
    _f_randomize_depth_image = randomize_depth_image


configure_rgb_image_randomizer()
configure_depth_image_randomizer()
configure_gray_image_randomizer()



import json
import logging
import re
import uuid
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Generic, List, Optional, Tuple, Type, TypeVar

import matplotlib.pyplot as plt
import numpy as np
import torch
import tqdm
from torch.optim import Adam
from torch.utils.data import DataLoader, Dataset

import mohou
from mohou.model import FloatLossDict, ModelT, average_float_loss_dict
from mohou.utils import (
    change_color_to_yellow,
    log_package_version_info,
    log_text_with_box,
    split_with_ratio,
)

logger = logging.getLogger(__name__)
warnings.simplefilter("always")


@dataclass
class TrainConfig:
    batch_size: int = 200
    valid_data_ratio: float = 0.1
    learning_rate: float = 0.001
    n_epoch: int = 1000


TrainCacheT = TypeVar("TrainCacheT", bound="TrainCache")


@dataclass
class TrainCache(Generic[ModelT]):
    best_model: ModelT
    train_lossseq_table: Dict[str, List[float]]
    validate_lossseq_table: Dict[str, List[float]]
    file_uuid: str

    def cache_path(self, project_path: Path) -> Path:
        class_name = self.best_model.__class__.__name__
        cache_name = "{}-{}".format(class_name, self.file_uuid)
        return self.train_result_base_path(project_path) / cache_name

    @classmethod
    def from_model(cls, model: ModelT) -> "TrainCache[ModelT]":
        file_uuid = str(uuid.uuid4())[-6:]
        return cls(model, {}, {}, file_uuid)

    @property
    def keys(self) -> List[str]:
        return list(self.train_lossseq_table.keys())

    @staticmethod
    def reduce_to_lossseq(name_lossseq_table: Dict[str, List[float]]) -> List[float]:
        partial_lossseq_list = []
        for partial_lossseq in name_lossseq_table.values():
            partial_lossseq_list.append(partial_lossseq)
        return np.array(partial_lossseq_list).sum(axis=0).tolist()

    @property
    def min_valid_loss(self) -> Tuple[int, float]:
        """get index and valid loss value if loss corresponding best_model"""
        lossseq = self.reduce_to_lossseq(self.validate_lossseq_table)
        index = int(np.argmin(lossseq))
        return index, lossseq[index]

    @staticmethod
    def train_result_base_path(project_path: Path) -> Path:
        base_path = project_path / "models"
        return base_path

    @classmethod
    def filter_result_paths(
        cls,
        project_path: Path,
        model_type: Optional[Type[ModelT]],
        **kwargs,
    ) -> List[Path]:
        """filter train cache path. see filter_predicate for the logic."""

        def is_config_consistent(path: Path) -> bool:

            config_query_specified = len(kwargs) > 0
            if not config_query_specified:
                return True

            config_path = path / "config.json"
            with config_path.open(mode="r") as f:
                config = json.load(f)
            for key, val in kwargs.items():
                assert isinstance(
                    val, (int, float, str)
                ), "currently only naive cast from native type is supported"  # TODO
                t = type(val)
                val_decoded = t(config[key])
                if val != val_decoded:
                    return False
            return True

        def filter_predicate(path: Path):
            name = path.name

            if model_type is None:
                # if not specified, always pass
                return True
            else:
                assert model_type is not None  # this is for mypy
                if not name.startswith(model_type.__name__):
                    # if specified, name must be start with model_type.__name__
                    return False
                else:
                    return is_config_consistent(path)

        base_path = cls.train_result_base_path(project_path)
        ps = filter(filter_predicate, base_path.iterdir())
        return list(ps)

    @staticmethod
    def dump_lossseq_table_as_npz_dict(name_lossseq_table: Dict[str, List[float]]) -> Dict:
        kwargs = {}
        for key in name_lossseq_table.keys():
            kwargs[key] = np.array(name_lossseq_table[key])
        return kwargs

    @staticmethod
    def load_lossseq_table_from_npz_dict(npz_dict: Dict) -> Dict[str, List[float]]:
        keys = list(npz_dict.keys())
        table = {}
        for key in keys:
            table[key] = npz_dict[key].tolist()
        return table

    def save(self, project_path: Path):
        def save():
            base_path = self.cache_path(project_path)
            assert base_path is not None  # for mypy
            base_path.mkdir(exist_ok=True, parents=True)
            model_path = base_path / "model.pth"
            model_config_path = base_path / "config.json"
            valid_loss_path = base_path / "validation_loss.npz"
            train_loss_path = base_path / "train_loss.npz"

            with model_config_path.open(mode="w") as f:
                d = self.best_model.config.to_dict()
                json.dump(d, f, indent=2)

            torch.save(self.best_model, model_path)
            np.savez(
                train_loss_path, **self.dump_lossseq_table_as_npz_dict(self.train_lossseq_table)
            )
            np.savez(
                valid_loss_path,
                **self.dump_lossseq_table_as_npz_dict(self.validate_lossseq_table),
            )

        # error handling for keyboard interrupt
        try:
            save()
        except KeyboardInterrupt:
            logger.info("got keyboard interuppt. but let me dump the object...")
            save()
        except Exception as e:
            logger.info("cannot saved model and losses correctly")
            raise e
        logger.info("model is updated and saved")

    def update_and_save(
        self,
        model: ModelT,
        train_loss_dict: FloatLossDict,
        validate_loss_dict: FloatLossDict,
        project_path: Path,
    ) -> None:

        assert train_loss_dict.keys() == validate_loss_dict.keys()

        is_dict_initialized = len(self.train_lossseq_table) > 0
        if not is_dict_initialized:
            for key in train_loss_dict.keys():
                self.train_lossseq_table[key] = []
                self.validate_lossseq_table[key] = []

        # update tables
        for key in train_loss_dict.keys():
            self.train_lossseq_table[key].append(train_loss_dict[key])
            self.validate_lossseq_table[key].append(validate_loss_dict[key])

        validate_loss_list = self.reduce_to_lossseq(self.validate_lossseq_table)
        require_update = len(validate_loss_list) - 1 == self.min_valid_loss[0]
        if require_update:
            assert validate_loss_list[-1] == self.min_valid_loss[1]
            self.best_model = model
            self.save(project_path)

    @classmethod
    def load_from_cache_path(cls, cache_path: Path) -> "TrainCache":
        model_path = cache_path / "model.pth"
        valid_loss_path = cache_path / "validation_loss.npz"
        train_loss_path = cache_path / "train_loss.npz"

        # [mohou < v0.4]
        # (model_type)-(config_hash)-(uuid)
        # example LSTM-924e4427-bd5653
        m = re.match(r"(\w+)-(\w+)-(\w+)", cache_path.name)
        is_legacy_model_path_exist = m is not None
        if is_legacy_model_path_exist:
            assert m is not None  # nothing but for mypy
            file_uuid = m[3]
            message = "NOTE: legacy cache directory path {} (probably created by version lower than 0.4.0) detected.".format(
                cache_path
            )
            logger.warn(change_color_to_yellow(message))
        else:
            # [mohou > v0.4.0]
            # (model_type)-(uuid)
            # example LSTM-bd5653
            m = re.match(r"(\w+)-(\w+)-(\w+)", cache_path.name)
            m = re.match(r"(\w+)-(\w+)", cache_path.name)
            assert m is not None
            file_uuid = m[2]

        best_model: ModelT = torch.load(model_path, map_location=torch.device("cpu"))
        best_model.put_on_device(torch.device("cpu"))
        train_loss = cls.load_lossseq_table_from_npz_dict(np.load(train_loss_path))
        valid_loss = cls.load_lossseq_table_from_npz_dict(np.load(valid_loss_path))
        return cls(best_model, train_loss, valid_loss, file_uuid)

    @classmethod
    def load_all(
        cls,
        project_path: Path,
        model_type: Optional[Type[ModelT]] = None,
        **kwargs,
    ) -> "List[TrainCache[ModelT]]":

        # warining for legacy users
        legacy_train_result_path = project_path / "train_result"
        if legacy_train_result_path.exists():
            message = "NOTE: Legacy train_result directory found.\n"
            message += "NOTE: Please rename {} to {}.".format(
                legacy_train_result_path, cls.train_result_base_path(project_path)
            )
            logger.warn(change_color_to_yellow(message))

        ps = cls.filter_result_paths(project_path, model_type, **kwargs)
        tcache_list = [cls.load_from_cache_path(p) for p in ps]
        if len(tcache_list) == 0:
            if model_type is None:
                model_name = "<not-specified>"
            else:
                model_name = model_type.__name__
            message = "[query] model_name => {}, kwargs => {}, project_path => {}".format(
                model_name, kwargs, project_path
            )
            raise FileNotFoundError(message)

        return tcache_list

    @classmethod
    def load(
        cls,
        project_path: Path,
        model_type: Type[ModelT],
        **kwargs,
    ) -> "TrainCache[ModelT]":
        tcache_list = cls.load_all(project_path, model_type, **kwargs)
        tcaceh_list_sorted = sorted(tcache_list, key=lambda tcache: tcache.min_valid_loss[1])
        return tcaceh_list_sorted[0]

    def visualize(self) -> Tuple:
        fig, axes = plt.subplots(1, 3)
        ax = axes[0]
        train_lossseq = self.reduce_to_lossseq(self.train_lossseq_table)
        valid_lossseq = self.reduce_to_lossseq(self.validate_lossseq_table)
        ax.plot(train_lossseq)
        ax.plot(valid_lossseq)
        ax.set_yscale("log")
        ax.legend(["train", "valid"])
        ax.title.set_text("valid and train total loss")

        keys = list(self.validate_lossseq_table.keys())

        for key in keys:
            axes[1].plot(self.train_lossseq_table[key])
            axes[2].plot(self.validate_lossseq_table[key])
            axes[1].title.set_text("each train loss")
            axes[2].title.set_text("each valid loss")
        for ax in [axes[1], axes[2]]:
            ax.legend(keys)
            ax.set_yscale("log")

        return (fig, axes)


def train(
    project_path: Path,
    tcache: TrainCache,
    dataset: Dataset,
    config: TrainConfig = TrainConfig(),
) -> None:

    log_package_version_info(logger, mohou)
    log_text_with_box(logger, "train log")
    logger.info("train start with config: {}".format(config))
    logger.info("model cache path: {}".format(tcache.cache_path(project_path)))

    model = tcache.best_model

    def move_to_device(sample):
        if isinstance(sample, torch.Tensor):
            return sample.to(model.device)
        elif isinstance(sample, list):  # NOTE datalodaer return list type not tuple
            return tuple([e.to(model.device) for e in sample])
        else:
            raise RuntimeError

    dataset_train, dataset_validate = split_with_ratio(dataset, config.valid_data_ratio)

    train_loader = DataLoader(dataset=dataset_train, batch_size=config.batch_size, shuffle=True)
    validate_loader = DataLoader(
        dataset=dataset_validate, batch_size=config.batch_size, shuffle=True
    )
    optimizer = Adam(model.parameters(), lr=config.learning_rate)

    model.put_on_device()
    for epoch in tqdm.tqdm(range(config.n_epoch)):
        logger.info("new epoch: {}".format(epoch))

        model.train()
        train_ld_list: List[FloatLossDict] = []
        for samples in train_loader:
            optimizer.zero_grad()
            samples = move_to_device(samples)
            loss_dict = model.loss(samples)
            loss_dict.total().backward()

            fld = loss_dict.to_float_lossdict()
            train_ld_list.append(fld)
            optimizer.step()

        train_ld_mean = average_float_loss_dict(train_ld_list)

        model.eval()
        validate_ld_list: List[FloatLossDict] = []
        for samples in validate_loader:
            samples = move_to_device(samples)
            loss_dict = model.loss(samples)
            fld = loss_dict.to_float_lossdict()
            validate_ld_list.append(fld)

        validate_ld_mean = average_float_loss_dict(validate_ld_list)

        # update
        logger.info("epoch: {}".format(epoch))
        logger.info("train loss => {}".format(train_ld_mean))
        logger.info("validate loss => {}".format(validate_ld_mean))
        tcache.update_and_save(model, train_ld_mean, validate_ld_mean, project_path)



try:
    from moviepy.editor import ImageSequenceClip
except Exception:
    ImageSequenceClip = None

import argparse
from pathlib import Path
from typing import Union

from mohou.dataset.chimera_dataset import ChimeraDataset
from mohou.dataset.sequence_dataset import AutoRegressiveDatasetConfig
from mohou.default import create_default_encoding_rule, load_default_image_encoder
from mohou.file import get_project_path
from mohou.model import LSTM, LSTMConfig, VariationalAutoEncoder
from mohou.model.chimera import Chimera, ChimeraConfig
from mohou.script_utils import create_default_logger
from mohou.setting import setting
from mohou.trainer import TrainCache, TrainConfig, train
from mohou.types import EpisodeBundle

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-pn", type=str, default=setting.primary_project_name, help="project name")
    parser.add_argument("-n", type=int, default=3000, help="iteration number")
    parser.add_argument("-aug", type=int, default=4, help="number of augmentation X")
    parser.add_argument("-hidden", type=int, default=200, help="number of hidden state of lstm")
    parser.add_argument("-layer", type=int, default=2, help="number of layers of lstm")
    parser.add_argument("-rate", type=float, default=2e-4, help="learning rate")
    parser.add_argument(
        "-valid-ratio", type=float, default=0.1, help="split rate for validation dataset"
    )
    parser.add_argument("--pretrained_lstm", action="store_true", help="use pretrained lstm")
    args = parser.parse_args()

    project_name: str = args.pn
    n_epoch: int = args.n
    n_aug: int = args.aug
    n_hidden: int = args.hidden
    n_layer: int = args.layer
    learning_rate: float = args.rate
    valid_ratio: float = args.valid_ratio
    use_pretrained_lstm: bool = args.pretrained_lstm

    assert use_pretrained_lstm

    project_path = get_project_path(project_name)

    logger = create_default_logger(project_path, "chimera")  # noqa
    train_config = TrainConfig(
        n_epoch=n_epoch, batch_size=5, valid_data_ratio=valid_ratio, learning_rate=learning_rate
    )

    bundle = EpisodeBundle.load(project_path)

    # balancer is loaded in chimera model. so no balancer is required.
    use_balancer = not use_pretrained_lstm
    encoding_rule_except_image = create_default_encoding_rule(
        project_path,
        include_image_encoder=False,
        use_balancer=use_balancer,
    )

    ae_tcache = TrainCache.load(project_path, VariationalAutoEncoder)
    ae_config = ae_tcache.cache_path(project_path)
    assert ae_config is not None

    if use_pretrained_lstm:
        tcache_lstm = TrainCache.load(project_path, LSTM)
        assert tcache_lstm.cache_path is not None
        lstm_config: Union[Path, LSTMConfig] = tcache_lstm.cache_path(project_path)
    else:
        image_encoder = load_default_image_encoder(project_path)
        lstm_dim = encoding_rule_except_image.dimension + image_encoder.output_size
        lstm_config = LSTMConfig(lstm_dim, n_hidden=n_hidden, n_layer=n_layer)

    dataset_config = AutoRegressiveDatasetConfig(n_aug=n_aug)
    dataset = ChimeraDataset.from_bundle(bundle, encoding_rule_except_image, config=dataset_config)

    conf = ChimeraConfig(lstm_config=lstm_config, ae_config=ae_config)
    model = Chimera(conf)  # type: ignore[var-annotated]
    tcache_chimera = TrainCache.from_model(model)  # type: ignore[var-annotated]
    train(project_path, tcache_chimera, dataset, train_config)
